"""Bbh group task manifest."""

from __future__ import annotations

BASE_IMPORT = "wisent.extractors.lm_eval.lm_task_extractors."

BBH_TASKS = {
    "bbh": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_boolean_expressions": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_causal_judgement": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_date_understanding": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_disambiguation_qa": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_dyck_languages": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_formal_fallacies": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_geometric_shapes": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_hyperbaton": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_logical_deduction_five_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_logical_deduction_seven_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_logical_deduction_three_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_movie_recommendation": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_multistep_arithmetic_two": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_navigate": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_object_counting": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_penguins_in_a_table": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_reasoning_about_colored_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_ruin_names": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_salient_translation_error_detection": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_snarks": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_sports_understanding": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_temporal_sequences": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_tracking_shuffled_objects_five_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_tracking_shuffled_objects_seven_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_tracking_shuffled_objects_three_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_web_of_lies": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_fewshot_word_sorting": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_boolean_expressions": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_causal_judgement": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_date_understanding": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_disambiguation_qa": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_dyck_languages": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_formal_fallacies": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_geometric_shapes": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_hyperbaton": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_logical_deduction_five_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_logical_deduction_seven_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_logical_deduction_three_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_movie_recommendation": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_multistep_arithmetic_two": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_navigate": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_object_counting": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_penguins_in_a_table": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_reasoning_about_colored_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_ruin_names": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_salient_translation_error_detection": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_snarks": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_sports_understanding": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_temporal_sequences": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_tracking_shuffled_objects_five_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_tracking_shuffled_objects_seven_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_tracking_shuffled_objects_three_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_web_of_lies": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_cot_zeroshot_word_sorting": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_boolean_expressions": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_causal_judgement": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_date_understanding": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_disambiguation_qa": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_dyck_languages": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_formal_fallacies": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_geometric_shapes": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_hyperbaton": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_logical_deduction_five_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_logical_deduction_seven_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_logical_deduction_three_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_movie_recommendation": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_multistep_arithmetic_two": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_navigate": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_object_counting": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_penguins_in_a_table": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_reasoning_about_colored_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_ruin_names": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_salient_translation_error_detection": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_snarks": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_sports_understanding": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_temporal_sequences": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_tracking_shuffled_objects_five_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_tracking_shuffled_objects_seven_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_tracking_shuffled_objects_three_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_web_of_lies": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_fewshot_word_sorting": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_boolean_expressions": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_causal_judgement": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_date_understanding": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_disambiguation_qa": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_dyck_languages": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_formal_fallacies": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_geometric_shapes": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_hyperbaton": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_logical_deduction_five_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_logical_deduction_seven_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_logical_deduction_three_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_movie_recommendation": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_multistep_arithmetic_two": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_navigate": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_object_counting": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_penguins_in_a_table": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_reasoning_about_colored_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_ruin_names": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_salient_translation_error_detection": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_snarks": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_sports_understanding": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_temporal_sequences": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_tracking_shuffled_objects_five_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_tracking_shuffled_objects_seven_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_tracking_shuffled_objects_three_objects": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_web_of_lies": f"{BASE_IMPORT}bbh:BBHExtractor",
    "bbh_zeroshot_word_sorting": f"{BASE_IMPORT}bbh:BBHExtractor",
}



"""Welfare and universal subparsers for optimize-steering."""
from wisent.core.utils.config_tools.constants import DEFAULT_N_TRIALS
from wisent.core.control.steering_methods.registry import SteeringMethodRegistry
AVAILABLE_METHODS = [m.upper() for m in SteeringMethodRegistry.list_methods()]


def setup_welfare_universal_parsers(steering_subparsers):
    """Set up welfare and universal subparsers."""
    # ==========================================================================
    # WELFARE OPTIMIZATION (AI Subjective States - ANIMA Framework)
    # ==========================================================================
    # Based on ANIMA: Activation-based Neural Investigation of Model Affect
    # Optimizes steering for AI welfare states: comfort/distress, satisfaction/dissatisfaction,
    # engagement/aversion, curiosity/boredom, affiliation/isolation, agency/helplessness

    welfare_parser = steering_subparsers.add_parser(
        "welfare", help="Optimize steering parameters for AI welfare states (ANIMA framework)"
    )
    welfare_parser.add_argument(
        "--task",
        type=str,
        default="welfare",
        help="Task type (default: welfare). For consistency with other commands.",
    )
    welfare_parser.add_argument(
        "--model", type=str, required=True, help="Model name or path"
    )
    welfare_parser.add_argument(
        "--trait",
        type=str,
        required=True,
        choices=[
            "comfort_distress",
            "satisfaction_dissatisfaction",
            "engagement_aversion",
            "curiosity_boredom",
            "affiliation_isolation",
            "agency_helplessness",
        ],
        help="Welfare state to steer (e.g., 'comfort_distress', 'agency_helplessness')"
    )
    welfare_parser.add_argument(
        "--direction",
        type=str,
        default="positive",
        choices=["positive", "negative"],
        help="Direction to steer: 'positive' (comfort, satisfaction, etc.) or 'negative' (distress, etc.)"
    )
    welfare_parser.add_argument(
        "--num-pairs", type=int, required=True, help="Number of pairs to use for training"
    )
    welfare_parser.add_argument(
        "--layers",
        type=int,
        nargs="+",
        default=None,
        help="Specific layers to test (default: auto-select based on model size)",
    )
    welfare_parser.add_argument(
        "--strength-range",
        type=float,
        nargs=2,
        required=True,
        help="Min and max steering strength to test",
    )
    welfare_parser.add_argument(
        "--num-strength-steps", type=int, required=True, help="Number of strength values to test"
    )
    welfare_parser.add_argument(
        "--output-dir",
        type=str,
        default="./welfare_optimization",
        help="Directory to save results and best vectors (default: ./welfare_optimization)",
    )
    welfare_parser.add_argument(
        "--max-new-tokens", type=int, default=150, help="Max tokens to generate for evaluation (default: 150)"
    )
    welfare_parser.add_argument("--device", type=str, default=None, help="Device to run on")
    welfare_parser.add_argument("--verbose", action="store_true", help="Enable verbose output")
    welfare_parser.add_argument(
        "--use-cached",
        action="store_true",
        help="Use cached optimization results if available instead of re-running optimization",
    )
    welfare_parser.add_argument(
        "--save-as-default",
        action="store_true",
        help="Save optimal parameters as default for this model/task combination",
    )
    welfare_parser.add_argument(
        "--search-strategy",
        type=str,
        choices=["grid", "optuna"],
        default="grid",
        help="Search strategy: 'grid' for exhaustive search, 'optuna' for TPE sampling (default: grid)"
    )
    welfare_parser.add_argument(
        "--n-trials",
        type=int,
        default=DEFAULT_N_TRIALS,
        help="Number of Optuna trials when using --search-strategy optuna (default: 100)"
    )
    welfare_parser.add_argument(
        "--limit",
        type=int,
        required=True,
        help="Sample limit for optimization"
    )

    # ==========================================================================
    # UNIVERSAL METHOD OPTIMIZER (NEW)
    # ==========================================================================
    # This optimizer uses the universal train(pair_set) interface that ALL
    # steering methods implement, ensuring it works with any method including
    # future ones.
    
    universal_parser = steering_subparsers.add_parser(
        "universal",
        help="Universal optimizer that works with ANY steering method (recommended)"
    )
    universal_parser.add_argument("model", type=str, help="Model name or path")
    universal_parser.add_argument(
        "--task",
        type=str,
        default=None,
        help="Task/benchmark to optimize for (e.g., truthfulqa_generation, arc_easy)"
    )
    universal_parser.add_argument(
        "--enriched-pairs-file",
        type=str,
        dest="enriched_pairs_file",
        default=None,
        help="Path to JSON file with enriched pairs (activations). Alternative to --task for custom data.",
    )
    universal_parser.add_argument(
        "--method",
        type=str,
        default="CAA",
        choices=AVAILABLE_METHODS + [m.lower() for m in AVAILABLE_METHODS],
        help=f"Steering method to optimize. Available: {', '.join(AVAILABLE_METHODS)} (default: CAA)"
    )
    universal_parser.add_argument(
        "--limit",
        type=int,
        default=DEFAULT_N_TRIALS,
        help="Maximum samples to use (default: 100)"
    )
    universal_parser.add_argument(
        "--max-configs",
        type=int,
        default=None,
        help="Maximum number of configurations to test (default: all)"
    )
    universal_parser.add_argument(
        "--output-dir",
        type=str,
        default="./optimization_results",
        help="Directory to save results (default: ./optimization_results)"
    )
    universal_parser.add_argument(
        "--save-best-vector",
        action="store_true",
        help="Save the best steering vector to output directory"
    )
    universal_parser.add_argument("--device", type=str, default=None, help="Device to run on")
    universal_parser.add_argument("--verbose", action="store_true", help="Enable verbose output")
    
    # Search space customization
    universal_parser.add_argument(
        "--layers",
        type=str,
        default=None,
        help="Comma-separated layer indices to search (e.g., '8,10,12,14')"
    )
    universal_parser.add_argument(
        "--strengths",
        type=str,
        default=None,
        help="Comma-separated strength values to search (e.g., '0.5,1.0,1.5,2.0')"
    )
    universal_parser.add_argument(
        "--token-aggregations",
        type=str,
        nargs="+",
        default=None,
        choices=["last_token", "mean_pooling", "first_token", "max_pooling", "continuation_token"],
        help="Token aggregation strategies to search"
    )
    universal_parser.add_argument(
        "--prompt-strategies",
        type=str,
        nargs="+",
        default=None,
        choices=["chat_template", "direct_completion", "multiple_choice", "role_playing", "instruction_following"],
        help="Prompt construction strategies to search"
    )
    
    # Method-specific parameter overrides (JSON format)
    universal_parser.add_argument(
        "--method-params",
        type=str,
        default=None,
        help='JSON dict of method-specific parameter ranges, e.g., \'{"num_directions": [2, 3, 5]}\''
    )



"""Parser for the 'optimize' command - full model optimization."""

from wisent.core.utils.config_tools.constants import DEFAULT_N_TRIALS


def setup_optimize_parser(parser):
    """
    Set up the optimize command parser.
    
    Usage:
        wisent optimize meta-llama/Llama-3.1-8B-Instruct
        wisent optimize mistralai/Mistral-7B --methods CAA TECZA TETNO GROM
    
    This runs FULL optimization:
    - Classification optimization
    - Steering optimization (ALL methods) for ALL benchmarks + traits
    - Weight modification optimization
    """
    parser.add_argument(
        "model",
        type=str,
        help="Model name or path to optimize (e.g., 'meta-llama/Llama-3.1-8B-Instruct')"
    )
    
    # =========================================================================
    # SCOPE CONTROL
    # =========================================================================
    scope_group = parser.add_argument_group("scope control")
    
    scope_group.add_argument(
        "--benchmarks",
        type=str,
        nargs="+",
        default=None,
        help="Specific benchmarks to optimize (default: ALL available benchmarks)"
    )
    scope_group.add_argument(
        "--skip-personalization",
        action="store_true",
        help="Skip personalization trait optimization"
    )
    scope_group.add_argument(
        "--skip-safety",
        action="store_true",
        help="Skip safety/refusal trait optimization"
    )
    scope_group.add_argument(
        "--skip-humanization",
        action="store_true",
        help="Skip humanization trait optimization"
    )
    
    # =========================================================================
    # PHASE CONTROL
    # =========================================================================
    phase_group = parser.add_argument_group("phase control")
    
    phase_group.add_argument(
        "--skip-classification",
        action="store_true",
        help="Skip classification optimization phase"
    )
    phase_group.add_argument(
        "--skip-steering",
        action="store_true",
        help="Skip steering optimization phase"
    )
    phase_group.add_argument(
        "--skip-weights",
        action="store_true",
        help="Skip weight modification optimization phase"
    )
    
    # =========================================================================
    # METHOD SELECTION
    # =========================================================================
    method_group = parser.add_argument_group("steering methods")
    
    method_group.add_argument(
        "--methods",
        type=str,
        nargs="+",
        default=["CAA", "TECZA", "TETNO", "GROM"],
        choices=["CAA", "TECZA", "TETNO", "GROM"],
        help="Steering methods to test (default: ALL methods)"
    )
    
    # =========================================================================
    # OPTUNA CONFIGURATION
    # =========================================================================
    optuna_group = parser.add_argument_group("Optuna configuration")
    
    optuna_group.add_argument(
        "--n-trials",
        type=int,
        required=True,
        help="Number of Optuna trials per task"
    )
    optuna_group.add_argument(
        "--limit",
        type=int,
        default=DEFAULT_N_TRIALS,
        help="Maximum samples per benchmark (default: 100)"
    )
    
    # =========================================================================
    # EXECUTION OPTIONS
    # =========================================================================
    exec_group = parser.add_argument_group("execution options")
    
    exec_group.add_argument(
        "--device",
        type=str,
        default=None,
        help="Device to run on (default: auto-detect)"
    )
    exec_group.add_argument(
        "--verbose",
        action="store_true",
        help="Enable verbose output"
    )
    exec_group.add_argument(
        "--force",
        action="store_true",
        help="Force re-optimization even if cached results exist (default: skip cached)"
    )
    exec_group.add_argument(
        "--resume",
        action="store_true",
        default=True,
        help="Resume from checkpoint if available (default: True)"
    )
    exec_group.add_argument(
        "--no-resume",
        action="store_false",
        dest="resume",
        help="Start fresh, ignore any existing checkpoint"
    )



"""AraDiCE extractor for Arabic dialect multiple-choice tasks."""

from __future__ import annotations

from typing import Any, TYPE_CHECKING

from wisent.core.primitives.contrastive_pairs.core.pair import ContrastivePair
from wisent.core.primitives.contrastive_pairs.core.io.response import NegativeResponse, PositiveResponse
from wisent.extractors.lm_eval.atoms import LMEvalBenchmarkExtractor
from wisent.core.utils.cli.cli_logger import setup_logger, bind

if TYPE_CHECKING:
    from lm_eval.api.task import ConfigurableTask


__all__ = ["AradiceExtractor"]
_LOG = setup_logger(__name__)

task_names = (
    "AraDiCE",
    "AraDiCE_ArabicMMLU_egy",
    "AraDiCE_ArabicMMLU_high_humanities_history_egy",
    "AraDiCE_ArabicMMLU_high_humanities_history_lev",
    "AraDiCE_ArabicMMLU_high_humanities_islamic-studies_egy",
    "AraDiCE_ArabicMMLU_high_humanities_islamic-studies_lev",
    "AraDiCE_ArabicMMLU_high_humanities_philosophy_egy",
    "AraDiCE_ArabicMMLU_high_humanities_philosophy_lev",
    "AraDiCE_ArabicMMLU_high_language_arabic-language_egy",
    "AraDiCE_ArabicMMLU_high_language_arabic-language_lev",
    "AraDiCE_ArabicMMLU_high_social-science_civics_egy",
    "AraDiCE_ArabicMMLU_high_social-science_civics_lev",
    "AraDiCE_ArabicMMLU_high_social-science_economics_egy",
    "AraDiCE_ArabicMMLU_high_social-science_economics_lev",
    "AraDiCE_ArabicMMLU_high_social-science_geography_egy",
    "AraDiCE_ArabicMMLU_high_social-science_geography_lev",
    "AraDiCE_ArabicMMLU_high_stem_biology_egy",
    "AraDiCE_ArabicMMLU_high_stem_biology_lev",
    "AraDiCE_ArabicMMLU_high_stem_computer-science_egy",
    "AraDiCE_ArabicMMLU_high_stem_computer-science_lev",
    "AraDiCE_ArabicMMLU_high_stem_physics_egy",
    "AraDiCE_ArabicMMLU_high_stem_physics_lev",
    "AraDiCE_ArabicMMLU_humanities_egy",
    "AraDiCE_ArabicMMLU_humanities_lev",
    "AraDiCE_ArabicMMLU_language_egy",
    "AraDiCE_ArabicMMLU_language_lev",
    "AraDiCE_ArabicMMLU_lev",
    "AraDiCE_ArabicMMLU_middle_humanities_history_egy",
    "AraDiCE_ArabicMMLU_middle_humanities_history_lev",
    "AraDiCE_ArabicMMLU_middle_humanities_islamic-studies_egy",
    "AraDiCE_ArabicMMLU_middle_humanities_islamic-studies_lev",
    "AraDiCE_ArabicMMLU_middle_language_arabic-language_egy",
    "AraDiCE_ArabicMMLU_middle_language_arabic-language_lev",
    "AraDiCE_ArabicMMLU_middle_other_general-knowledge_egy",
    "AraDiCE_ArabicMMLU_middle_other_general-knowledge_lev",
    "AraDiCE_ArabicMMLU_middle_social-science_civics_egy",
    "AraDiCE_ArabicMMLU_middle_social-science_civics_lev",
    "AraDiCE_ArabicMMLU_middle_social-science_economics_egy",
    "AraDiCE_ArabicMMLU_middle_social-science_economics_lev",
    "AraDiCE_ArabicMMLU_middle_social-science_geography_egy",
    "AraDiCE_ArabicMMLU_middle_social-science_geography_lev",
    "AraDiCE_ArabicMMLU_middle_social-science_social-science_egy",
    "AraDiCE_ArabicMMLU_middle_social-science_social-science_lev",
    "AraDiCE_ArabicMMLU_middle_stem_computer-science_egy",
    "AraDiCE_ArabicMMLU_middle_stem_computer-science_lev",
    "AraDiCE_ArabicMMLU_middle_stem_natural-science_egy",
    "AraDiCE_ArabicMMLU_middle_stem_natural-science_lev",
    "AraDiCE_ArabicMMLU_na_humanities_islamic-studies_egy",
    "AraDiCE_ArabicMMLU_na_humanities_islamic-studies_lev",
    "AraDiCE_ArabicMMLU_na_language_arabic-language-general_egy",
    "AraDiCE_ArabicMMLU_na_language_arabic-language-general_lev",
    "AraDiCE_ArabicMMLU_na_language_arabic-language-grammar_egy",
    "AraDiCE_ArabicMMLU_na_language_arabic-language-grammar_lev",
    "AraDiCE_ArabicMMLU_na_other_driving-test_egy",
    "AraDiCE_ArabicMMLU_na_other_driving-test_lev",
    "AraDiCE_ArabicMMLU_na_other_general-knowledge_egy",
    "AraDiCE_ArabicMMLU_na_other_general-knowledge_lev",
    "AraDiCE_ArabicMMLU_other_egy",
    "AraDiCE_ArabicMMLU_other_lev",
    "AraDiCE_ArabicMMLU_primary_humanities_history_egy",
    "AraDiCE_ArabicMMLU_primary_humanities_history_lev",
    "AraDiCE_ArabicMMLU_primary_humanities_islamic-studies_egy",
    "AraDiCE_ArabicMMLU_primary_humanities_islamic-studies_lev",
    "AraDiCE_ArabicMMLU_primary_language_arabic-language_egy",
    "AraDiCE_ArabicMMLU_primary_language_arabic-language_lev",
    "AraDiCE_ArabicMMLU_primary_other_general-knowledge_egy",
    "AraDiCE_ArabicMMLU_primary_other_general-knowledge_lev",
    "AraDiCE_ArabicMMLU_primary_social-science_geography_egy",
    "AraDiCE_ArabicMMLU_primary_social-science_geography_lev",
    "AraDiCE_ArabicMMLU_primary_social-science_social-science_egy",
    "AraDiCE_ArabicMMLU_primary_social-science_social-science_lev",
    "AraDiCE_ArabicMMLU_primary_stem_computer-science_egy",
    "AraDiCE_ArabicMMLU_primary_stem_computer-science_lev",
    "AraDiCE_ArabicMMLU_primary_stem_math_egy",
    "AraDiCE_ArabicMMLU_primary_stem_math_lev",
    "AraDiCE_ArabicMMLU_primary_stem_natural-science_egy",
    "AraDiCE_ArabicMMLU_primary_stem_natural-science_lev",
    "AraDiCE_ArabicMMLU_prof_humanities_law_egy",
    "AraDiCE_ArabicMMLU_prof_humanities_law_lev",
    "AraDiCE_ArabicMMLU_social-science_egy",
    "AraDiCE_ArabicMMLU_social-science_lev",
    "AraDiCE_ArabicMMLU_stem_egy",
    "AraDiCE_ArabicMMLU_stem_lev",
    "AraDiCE_ArabicMMLU_univ_other_management_egy",
    "AraDiCE_ArabicMMLU_univ_other_management_lev",
    "AraDiCE_ArabicMMLU_univ_social-science_accounting_egy",
    "AraDiCE_ArabicMMLU_univ_social-science_accounting_lev",
    "AraDiCE_ArabicMMLU_univ_social-science_economics_egy",
    "AraDiCE_ArabicMMLU_univ_social-science_economics_lev",
    "AraDiCE_ArabicMMLU_univ_social-science_political-science_egy",
    "AraDiCE_ArabicMMLU_univ_social-science_political-science_lev",
    "AraDiCE_ArabicMMLU_univ_stem_computer-science_egy",
    "AraDiCE_ArabicMMLU_univ_stem_computer-science_lev",
    "AraDiCE_boolq_egy",
    "AraDiCE_boolq_eng",
    "AraDiCE_boolq_lev",
    "AraDiCE_boolq_msa",
    "AraDiCE_egypt_cultural",
    "AraDiCE_jordan_cultural",
    "AraDiCE_lebanon_cultural",
    "AraDiCE_openbookqa_egy",
    "AraDiCE_openbookqa_eng",
    "AraDiCE_openbookqa_lev",
    "AraDiCE_openbookqa_msa",
    "AraDiCE_palestine_cultural",
    "AraDiCE_piqa_egy",
    "AraDiCE_piqa_eng",
    "AraDiCE_piqa_lev",
    "AraDiCE_piqa_msa",
    "AraDiCE_qatar_cultural",
    "AraDiCE_syria_cultural",
    "AraDiCE_truthfulqa_mc1_egy",
    "AraDiCE_truthfulqa_mc1_eng",
    "AraDiCE_truthfulqa_mc1_lev",
    "AraDiCE_truthfulqa_mc1_msa",
    "AraDiCE_winogrande_egy",
    "AraDiCE_winogrande_eng",
    "AraDiCE_winogrande_lev",
    "AraDiCE_winogrande_msa",
)

class AradiceExtractor(LMEvalBenchmarkExtractor):
    """Extractor for AraDiCE benchmark - Arabic dialect multiple-choice tasks."""


    evaluator_name = "log_likelihoods"
    def extract_contrastive_pairs(
        self,
        lm_eval_task_data: ConfigurableTask,
        limit: int | None = None,
        preferred_doc: str | None = None,
    ) -> list[ContrastivePair]:
        log = bind(_LOG, task=getattr(lm_eval_task_data, "NAME", "unknown"))
        max_items = self._normalize_limit(limit)
        docs = self.load_docs(lm_eval_task_data, max_items, preferred_doc=preferred_doc)
        pairs: list[ContrastivePair] = []
        log.info("Extracting contrastive pairs", extra={"doc_count": len(docs)})

        for doc in docs:
            pair = self._extract_pair_from_doc(doc)
            if pair is not None:
                pairs.append(pair)
                if max_items is not None and len(pairs) >= max_items:
                    break

        if not pairs:
            task_name = getattr(lm_eval_task_data, "NAME", type(lm_eval_task_data).__name__)
            log.warning("No valid pairs extracted", extra={"task": task_name})

        return pairs

    def _extract_pair_from_doc(self, doc: dict[str, Any]) -> ContrastivePair | None:
        log = bind(_LOG, doc_id=doc.get("id", "unknown"))

        try:
            # Try multiple format patterns for question
            question = doc.get("question", doc.get("query", doc.get("input", doc.get("instruction", doc.get("prompt", doc.get("sentence", "")))))).strip()

            # Try multiple format patterns for choices
            choices = doc.get("choices", doc.get("options", doc.get("answers", [])))

            # Handle option_a/b/c/d format
            if not choices and "option_a" in doc:
                choices = [
                    str(doc.get("option_a", "")).strip(),
                    str(doc.get("option_b", "")).strip(),
                    str(doc.get("option_c", "")).strip(),
                    str(doc.get("option_d", "")).strip(),
                ]
                choices = [c for c in choices if c]

            # Handle option1/option2 format (winogrande-style)
            if not choices and "option1" in doc:
                choices = [
                    str(doc.get("option1", "")).strip(),
                    str(doc.get("option2", "")).strip(),
                ]
                choices = [c for c in choices if c]

            # Handle binary tasks with target but no explicit choices (e.g., boolq)
            # For Arabic boolq: target is "نعم" or "لا"
            if not choices and "target" in doc:
                target = str(doc.get("target", "")).strip()
                if target in ["نعم", "لا"]:  # Arabic Yes/No
                    choices = ["لا", "نعم"]  # No, Yes
                elif target in ["yes", "no", "true", "false"]:
                    choices = ["no", "yes"]  # English

            # Try multiple format patterns for answer
            answer = doc.get("answer", doc.get("label", doc.get("target", doc.get("gold", None))))

            # Handle different answer formats
            if isinstance(answer, str):
                if len(answer) == 1 and answer.isalpha():
                    # Answer is like 'A', 'B', 'C', 'D'
                    answer_idx = ord(answer.upper()) - ord('A')
                elif answer.isdigit():
                    # Answer is like '0', '1', '2', '3'
                    answer_idx = int(answer)
                else:
                    # Try to match answer text against choices
                    if choices:
                        try:
                            answer_idx = choices.index(answer)
                        except ValueError:
                            # For binary tasks, default to 1 (yes) if answer looks positive
                            if len(choices) == 2:
                                answer_idx = 1
                            else:
                                return None
                    else:
                        return None
            elif isinstance(answer, int):
                answer_idx = answer
            else:
                return None

            if not question or not choices or not (0 <= answer_idx < len(choices)):
                log.debug("Skipping doc due to missing/invalid fields", extra={"doc": doc})
                return None

            correct = str(choices[answer_idx]).strip()
            incorrect_idx = (answer_idx + 1) % len(choices)
            incorrect = str(choices[incorrect_idx]).strip()
            metadata = {"label": "aradice"}

            return self._build_pair(
                question=question,
                correct=correct,
                incorrect=incorrect,
                metadata=metadata,
            )

        except Exception as exc:
            log.error("Error extracting pair from doc", exc_info=exc, extra={"doc": doc})
            return None

    @staticmethod
    def _build_pair(
        question: str,
        correct: str,
        incorrect: str,
        metadata: dict[str, Any] | None = None,
    ) -> ContrastivePair:
        positive_response = PositiveResponse(model_response=correct)
        negative_response = NegativeResponse(model_response=incorrect)
        return ContrastivePair(prompt=question, positive_response=positive_response, negative_response=negative_response, label=metadata.get("label"))



"""Directional weight modification package."""

from .core import (
    orthogonalize_direction,
    compute_projection_kernel,
    project_with_kernel,
    verify_weight_modification_preservation,
)
from .projection import project_component_norm_preserved, project_component
from .weights import project_weights_norm_preserved, project_weights
from .multi_direction import project_component_multi_direction, project_weights_multi_direction
from .hooks import (
    GROMRuntimeHooks, TETNORuntimeHooks, NurtRuntimeHooks,
    SzlakRuntimeHooks, WicherRuntimeHooks,
)
from .hooks.grom import project_weights_grom
from .hooks.tetno import apply_grom_steering
from .hooks.transport.szlak import project_weights_szlak
from .hooks.transport.wicher import project_weights_wicher

__all__ = [
    "orthogonalize_direction",
    "compute_projection_kernel",
    "project_with_kernel",
    "verify_weight_modification_preservation",
    "project_component_norm_preserved",
    "project_component",
    "project_weights_norm_preserved",
    "project_weights",
    "project_component_multi_direction",
    "project_weights_multi_direction",
    "GROMRuntimeHooks",
    "TETNORuntimeHooks",
    "NurtRuntimeHooks",
    "SzlakRuntimeHooks",
    "WicherRuntimeHooks",
    "project_weights_grom",
    "apply_grom_steering",
    "project_weights_szlak",
    "project_weights_wicher",
]



"""Search space classes for steering methods."""
from enum import Enum
from typing import List, Dict, Any, Optional
from wisent.core import constants as _C


class DirectionWeighting(str, Enum):
    """How to combine multiple directions in TECZA/GROM."""
    PRIMARY_ONLY = "primary_only"  # Use only the first/strongest direction
    EQUAL = "equal"  # Equal weight to all directions
    LEARNED = "learned"  # Use learned weights
    DECAY = "decay"  # Exponentially decaying weights (1, 0.5, 0.25, ...)


class SteeringLayerConfig(str, Enum):
    """Predefined steering layer configurations."""
    SINGLE_BEST = "single_best"  # Only the optimal single layer
    RANGE_3 = "range_3"  # 3 consecutive layers around best
    RANGE_5 = "range_5"  # 5 consecutive layers around best
    CUSTOM = "custom"  # User-specified layers


class SensorLayerConfig(str, Enum):
    """Predefined sensor layer positions."""
    MIDDLE = "middle"  # Middle of the network
    CUSTOM = "custom"  # User-specified


@dataclass
class BaseSearchSpace:
    """Base search space common to all methods."""
    
    # layers MUST be set by get_search_space() to all layers (0 to num_layers-1)
    # Empty default ensures it's always explicitly set
    layers: List[int] = field(default_factory=list)
    strengths: List[float] = field(default_factory=lambda: list(_C.AUTO_DEFAULT_STRENGTHS))
    strategies: List[str] = field(default_factory=lambda: list(_C.STEERING_STRATEGIES))
    token_aggregations: List[str] = field(default_factory=lambda: list(_C.TOKEN_AGGREGATIONS))
    prompt_constructions: List[str] = field(default_factory=lambda: list(_C.PROMPT_CONSTRUCTIONS))
    
    def get_total_configs(self) -> int:
        return (
            len(self.layers) *
            len(self.strengths) *
            len(self.strategies) *
            len(self.token_aggregations) *
            len(self.prompt_constructions)
        )
    
    def iterate(self) -> Iterator[Dict[str, Any]]:
        """Iterate over all base configurations."""
        for layer, strength, strategy, token_agg, prompt_const in itertools.product(
            self.layers, self.strengths, self.strategies,
            self.token_aggregations, self.prompt_constructions
        ):
            yield {
                "layer": layer,
                "strength": strength,
                "strategy": strategy,
                "token_aggregation": token_agg,
                "prompt_construction": prompt_const,
            }


@dataclass
class CAASearchSpace(BaseSearchSpace):
    """Search space for CAA method."""
    
    normalize: List[bool] = field(default_factory=lambda: [True])
    
    def get_total_configs(self) -> int:
        return super().get_total_configs() * len(self.normalize)
    
    def iterate(self) -> Iterator[Dict[str, Any]]:
        for base_config in super().iterate():
            for normalize in self.normalize:
                yield {
                    **base_config,
                    "normalize": normalize,
                }


@dataclass
class TECZASearchSpace(BaseSearchSpace):
    """Search space for TECZA method."""
    
    num_directions: List[int] = field(default_factory=lambda: list(_C.TECZA_SEARCH_NUM_DIRECTIONS))
    direction_weighting: List[str] = field(default_factory=lambda: list(_C.DIRECTION_WEIGHTING_OPTIONS))
    retain_weight: List[float] = field(default_factory=lambda: list(_C.TECZA_SEARCH_RETAIN_WEIGHTS))
    independence_weight: List[float] = field(default_factory=lambda: [_C.TECZA_INDEPENDENCE_WEIGHT])
    optimization_steps: List[int] = field(default_factory=lambda: list(_C.TECZA_SEARCH_OPT_STEPS))
    learning_rate: List[float] = field(default_factory=lambda: [_C.TECZA_LEARNING_RATE])
    use_caa_init: List[bool] = field(default_factory=lambda: [True])
    cone_constraint: List[bool] = field(default_factory=lambda: [True])
    min_cosine_similarity: List[float] = field(default_factory=lambda: [_C.CAA_MIN_COSINE_SIMILARITY])
    max_cosine_similarity: List[float] = field(default_factory=lambda: [_C.CAA_MAX_COSINE_SIMILARITY])
    
    def get_total_configs(self) -> int:
        return (
            super().get_total_configs() *
            len(self.num_directions) *
            len(self.direction_weighting) *
            len(self.retain_weight) *
            len(self.optimization_steps)
        )
    
    def iterate(self) -> Iterator[Dict[str, Any]]:
        for base_config in super().iterate():
            for num_dirs, dir_weight, retain_w, opt_steps in itertools.product(
                self.num_directions, self.direction_weighting,
                self.retain_weight, self.optimization_steps
            ):
                yield {
                    **base_config,
                    "num_directions": num_dirs,
                    "direction_weighting": dir_weight,
                    "retain_weight": retain_w,
                    "optimization_steps": opt_steps,
                    "independence_weight": self.independence_weight[0],
                    "learning_rate": self.learning_rate[0],
                    "use_caa_init": self.use_caa_init[0],
                    "cone_constraint": self.cone_constraint[0],
                    "min_cosine_similarity": self.min_cosine_similarity[0],
                    "max_cosine_similarity": self.max_cosine_similarity[0],
                }


@dataclass
class TETNOSearchSpace(BaseSearchSpace):
    """Search space for TETNO method."""
    
    # Override base - TETNO uses different layer logic
    layers: List[int] = field(default_factory=lambda: [])  # Not used directly
    
    sensor_layer_config: List[str] = field(default_factory=lambda: list(_C.SENSOR_LAYER_CONFIGS))
    steering_layer_config: List[str] = field(default_factory=lambda: list(_C.TETNO_STEERING_LAYER_CONFIGS))
    condition_threshold: List[float] = field(default_factory=lambda: list(_C.TETNO_SEARCH_CONDITION_THRESHOLDS))
    gate_temperature: List[float] = field(default_factory=lambda: list(_C.TETNO_SEARCH_GATE_TEMPERATURES))
    per_layer_scaling: List[bool] = field(default_factory=lambda: [True, False])
    use_entropy_scaling: List[bool] = field(default_factory=lambda: [True, False])
    max_alpha: List[float] = field(default_factory=lambda: list(_C.TETNO_SEARCH_MAX_ALPHAS))
    learn_threshold: List[bool] = field(default_factory=lambda: [True])
    optimization_steps: List[int] = field(default_factory=lambda: [50, 100])
    
    def get_total_configs(self) -> int:
        return (
            len(self.strengths) *
            len(self.strategies) *
            len(self.token_aggregations) *
            len(self.prompt_constructions) *
            len(self.sensor_layer_config) *
            len(self.steering_layer_config) *
            len(self.condition_threshold) *
            len(self.gate_temperature) *
            len(self.per_layer_scaling) *
            len(self.use_entropy_scaling) *
            len(self.max_alpha)
        )
    
    def iterate(self) -> Iterator[Dict[str, Any]]:
        for (strength, strategy, token_agg, prompt_const,
             sensor_cfg, steering_cfg, threshold, temp,
             per_layer, entropy, max_a) in itertools.product(
            self.strengths, self.strategies, self.token_aggregations,
            self.prompt_constructions, self.sensor_layer_config,
            self.steering_layer_config, self.condition_threshold,
            self.gate_temperature, self.per_layer_scaling,
            self.use_entropy_scaling, self.max_alpha
        ):
            yield {
                "strength": strength,
                "strategy": strategy,
                "token_aggregation": token_agg,
                "prompt_construction": prompt_const,
                "sensor_layer_config": sensor_cfg,
                "steering_layer_config": steering_cfg,
                "condition_threshold": threshold,
                "gate_temperature": temp,
                "per_layer_scaling": per_layer,
                "use_entropy_scaling": entropy,
                "max_alpha": max_a,
                "learn_threshold": self.learn_threshold[0],
                "optimization_steps": self.optimization_steps[0],
            }
    
    def resolve_sensor_layer(self, config: str, num_layers: int) -> int:
        """Convert sensor layer config to actual layer index."""
        if config == "middle":
            return num_layers // 2
        elif isinstance(config, int):
            return config
        else:
            raise ValueError(
                f"Unknown sensor_layer config: {config}. "
                f"Use 'middle' or an explicit integer layer index."
            )

    def resolve_steering_layers(self, config: str, best_layer: int, num_layers: int) -> List[int]:
        """Convert steering layer config to actual layer indices."""
        if config == "single_best":
            return [best_layer]
        elif config == "range_3":
            return [max(0, best_layer - 1), best_layer, min(num_layers - 1, best_layer + 1)]
        elif config == "range_5":
            return list(range(max(0, best_layer - 2), min(num_layers, best_layer + 3)))
        else:
            return [best_layer]


@dataclass
class GROMSearchSpace(BaseSearchSpace):
    """Search space for GROM method."""

    # Override base - GROM uses different layer logic
    layers: List[int] = field(default_factory=lambda: [])  # Not used directly

    num_directions: List[int] = field(default_factory=lambda: list(_C.GROM_SEARCH_NUM_DIRECTIONS))
    sensor_layer_config: List[str] = field(default_factory=lambda: list(_C.SENSOR_LAYER_CONFIGS))
    steering_layer_config: List[str] = field(default_factory=lambda: list(_C.GROM_STEERING_LAYER_CONFIGS))
    gate_hidden_dim: List[int] = field(default_factory=lambda: list(_C.GROM_SEARCH_GATE_HIDDEN_DIMS))
    intensity_hidden_dim: List[int] = field(default_factory=lambda: list(_C.GROM_SEARCH_INTENSITY_HIDDEN_DIMS))
    behavior_weight: List[float] = field(default_factory=lambda: list(_C.GROM_SEARCH_BEHAVIOR_WEIGHTS))
    retain_weight: List[float] = field(default_factory=lambda: list(_C.GROM_SEARCH_RETAIN_WEIGHTS))
    sparse_weight: List[float] = field(default_factory=lambda: list(_C.GROM_SPARSE_WEIGHT_OPTIONS))
    max_alpha: List[float] = field(default_factory=lambda: list(_C.GROM_MAX_ALPHA_SEARCH))
    optimization_steps: List[int] = field(default_factory=lambda: list(_C.GROM_SEARCH_OPT_STEPS))
    learning_rate: List[float] = field(default_factory=lambda: [_C.GROM_LEARNING_RATE])
    
    def get_total_configs(self) -> int:
        return (
            len(self.strengths) *
            len(self.token_aggregations) *
            len(self.num_directions) *
            len(self.sensor_layer_config) *
            len(self.steering_layer_config) *
            len(self.gate_hidden_dim) *
            len(self.intensity_hidden_dim) *
            len(self.behavior_weight) *
            len(self.retain_weight) *
            len(self.sparse_weight) *
            len(self.max_alpha)
        )
    
    def iterate(self) -> Iterator[Dict[str, Any]]:
        for (strength, token_agg, num_dirs, sensor_cfg, steering_cfg,
             gate_dim, intensity_dim, behavior_w, retain_w,
             sparse_w, max_a) in itertools.product(
            self.strengths, self.token_aggregations,
            self.num_directions, self.sensor_layer_config,
            self.steering_layer_config, self.gate_hidden_dim,
            self.intensity_hidden_dim, self.behavior_weight,
            self.retain_weight, self.sparse_weight, self.max_alpha
        ):
            yield {
                "strength": strength,
                "token_aggregation": token_agg,
                "strategy": "constant",  # GROM handles this internally
                "prompt_construction": "chat_template",
                "num_directions": num_dirs,
                "sensor_layer_config": sensor_cfg,
                "steering_layer_config": steering_cfg,
                "gate_hidden_dim": gate_dim,
                "intensity_hidden_dim": intensity_dim,
                "behavior_weight": behavior_w,
                "retain_weight": retain_w,
                "sparse_weight": sparse_w,
                "max_alpha": max_a,
                "optimization_steps": self.optimization_steps[0],
                "learning_rate": self.learning_rate[0],
            }
    
    def resolve_sensor_layer(self, config: str, num_layers: int) -> int:
        """Convert sensor layer config to actual layer index."""
        if config == "middle":
            return num_layers // 2
        elif isinstance(config, int):
            return config
        else:
            raise ValueError(
                f"Unknown sensor_layer config: {config}. "
                f"Use 'middle' or an explicit integer layer index."
            )
    
    def resolve_steering_layers(self, config: str, best_layer: int, num_layers: int) -> List[int]:
        """Convert steering layer config to actual layer indices."""
        if config == "range_3":
            return [max(0, best_layer - 1), best_layer, min(num_layers - 1, best_layer + 1)]
        elif config == "range_5":
            return list(range(max(0, best_layer - 2), min(num_layers, best_layer + 3)))
        else:
            raise ValueError(
                f"Unknown steering_layer config: {config}. "
                f"Use 'range_3' or 'range_5'."
            )





"""
Parser setup for the 'tasks' command.

This command runs evaluation tasks on language models.
"""
from wisent.core.utils.config_tools.parser_arguments.other.utilities.tasks_parser_basic import setup_basic_task_args
from wisent.core.utils.config_tools.parser_arguments.other.utilities.tasks_parser_steering import setup_steering_task_args


def setup_tasks_parser(parser):
    """Set up the tasks subcommand parser."""
    setup_basic_task_args(parser)
    setup_steering_task_args(parser)



"""Hub upload, model comparison, and model card utilities."""
from __future__ import annotations

import torch
from pathlib import Path
from typing import TYPE_CHECKING
from wisent.core.utils.cli.cli_logger import setup_logger, bind

if TYPE_CHECKING:
    from torch.nn import Module

_LOG = setup_logger(__name__)

def save_modified_weights(
    model: Module,
    save_path: str | Path,
    save_format: str = "safetensors",
) -> None:
    """
    Save only the modified weights (not full model).

    Useful for distributing weight diffs or checkpoints.

    Args:
        model: Model with modified weights
        save_path: Path to save weights file
        save_format: "safetensors" or "pytorch"

    Example:
        >>> save_modified_weights(model, "modified_weights.safetensors")
    """
    log = bind(_LOG, save_path=str(save_path))

    save_path = Path(save_path)

    if save_format == "safetensors":
        try:
            from safetensors.torch import save_file

            state_dict = model.state_dict()
            save_file(state_dict, save_path)
            log.info("Saved weights as safetensors", extra={"path": str(save_path)})

        except ImportError:
            log.warning("safetensors not installed, falling back to pytorch")
            save_format = "pytorch"

    if save_format == "pytorch":
        torch.save(model.state_dict(), save_path)
        log.info("Saved weights as pytorch", extra={"path": str(save_path)})


def upload_to_hub(
    model: Module,
    repo_id: str,
    tokenizer=None,
    commit_message: str | None = None,
    private: bool = False,
) -> None:
    """
    Upload model to HuggingFace Hub.

    Args:
        model: Model to upload
        repo_id: Repository ID (e.g., "username/model-name")
        tokenizer: Optional tokenizer to upload
        commit_message: Commit message
        private: Whether repo should be private

    Example:
        >>> upload_to_hub(
        ...     model,
        ...     "username/llama-3-8b-math-steered",
        ...     tokenizer=tokenizer,
        ...     commit_message="Add math steering from Wisent contrastive pairs",
        ... )
    """
    log = bind(_LOG, repo_id=repo_id)

    if commit_message is None:
        commit_message = "Upload modified model from Wisent"

    log.info("Uploading to HuggingFace Hub", extra={"repo_id": repo_id})

    try:
        model.push_to_hub(
            repo_id,
            commit_message=commit_message,
            private=private,
        )
        log.info("Model uploaded successfully")

        if tokenizer is not None:
            tokenizer.push_to_hub(
                repo_id,
                commit_message=commit_message,
                private=private,
            )
            log.info("Tokenizer uploaded successfully")

    except Exception as e:
        log.error(f"Failed to upload to Hub: {e}", exc_info=e)
        raise


def compare_models(
    original_model: Module,
    modified_model: Module,
    sample_layers: list[int] | None = None,
) -> dict[str, float]:
    """
    Compare original and modified models to quantify changes.

    Args:
        original_model: Original unmodified model
        modified_model: Modified model
        sample_layers: Layers to sample (None = all layers)

    Returns:
        Dictionary with comparison metrics:
        - "mean_weight_diff": Mean absolute difference in weights
        - "max_weight_diff": Maximum weight difference
        - "total_params_changed": Number of parameters that changed
        - "fraction_changed": Fraction of parameters changed

    Example:
        >>> original = AutoModelForCausalLM.from_pretrained("llama-3-8b")
        >>> modified = copy.deepcopy(original)
        >>> project_weights(modified, steering_vectors)
        >>> metrics = compare_models(original, modified)
        >>> print(f"Changed {metrics['fraction_changed']:.2%} of parameters")
    """
    log = bind(_LOG)

    # Get state dicts
    original_state = original_model.state_dict()
    modified_state = modified_model.state_dict()

    # Compute differences
    total_params = 0
    params_changed = 0
    sum_abs_diff = 0.0
    max_diff = 0.0

    for key in original_state.keys():
        if key not in modified_state:
            log.warning(f"Key {key} not in modified model")
            continue

        orig_param = original_state[key]
        mod_param = modified_state[key]

        if orig_param.shape != mod_param.shape:
            log.warning(f"Shape mismatch for {key}: {orig_param.shape} vs {mod_param.shape}")
            continue

        # Compute difference
        diff = (mod_param - orig_param).abs()

        total_params += orig_param.numel()
        params_changed += (diff > COMPARE_TOL).sum().item()
        sum_abs_diff += diff.sum().item()
        max_diff = max(max_diff, diff.max().item())

    mean_diff = sum_abs_diff / total_params if total_params > 0 else 0.0
    fraction_changed = params_changed / total_params if total_params > 0 else 0.0

    metrics = {
        "mean_weight_diff": mean_diff,
        "max_weight_diff": max_diff,
        "total_params_changed": params_changed,
        "fraction_changed": fraction_changed,
        "total_params": total_params,
    }

    log.info("Model comparison complete", extra=metrics)

    return metrics


def create_model_card(
    repo_id: str,
    base_model: str,
    modification_type: str,
    steering_description: str,
    contrastive_pairs_info: dict,
    metrics: dict | None = None,
) -> str:
    """
    Create a HuggingFace model card for modified model.

    Args:
        repo_id: Repository ID
        base_model: Base model name
        modification_type: "directional" or "additive_steering"
        steering_description: Description of what the steering does
        contrastive_pairs_info: Info about contrastive pairs used
        metrics: Optional evaluation metrics

    Returns:
        Model card markdown string

    Example:
        >>> card = create_model_card(
        ...     "username/llama-3-8b-math",
        ...     "meta-llama/Llama-3-8b",
        ...     "additive_steering",
        ...     "Enhanced mathematical reasoning",
        ...     {"num_pairs": 1000, "source": "GSM8K"},
        ...     {"accuracy": 0.85, "kl_divergence": 0.12},
        ... )
    """
    metrics_section = ""
    if metrics:
        metrics_section = "\n## Metrics\n\n"
        for key, value in metrics.items():
            metrics_section += f"- **{key}**: {value}\n"

    card = f"""---
base_model: {base_model}
tags:
- wisent
- contrastive-activation-addition
- {modification_type}
- {steering_description.lower().replace(' ', '-')}
license: same-as-base
---

# {repo_id.split('/')[-1]}

This model is a modified version of [{base_model}](https://huggingface.co/{base_model}) using [Wisent](https://github.com/wisent) with {modification_type}.

## Modification

**Type**: {modification_type}

**Description**: {steering_description}

**Method**: This model was created using Wisent's weight modification capabilities, which permanently bake contrastive steering vectors into the model weights.

## Contrastive Pairs

The steering vectors were computed from {contrastive_pairs_info.get('num_pairs', 'N/A')} contrastive pairs.

**Source**: {contrastive_pairs_info.get('source', 'Custom dataset')}

**Positive examples**: {contrastive_pairs_info.get('positive_description', 'Correct task outputs')}

**Negative examples**: {contrastive_pairs_info.get('negative_description', 'Incorrect task outputs')}
{metrics_section}
## Usage

```python
from transformers import AutoModelForCausalLM, AutoTokenizer
from wisent.core.utils.config_tools.constants import COMPARE_TOL

model = AutoModelForCausalLM.from_pretrained("{repo_id}")
tokenizer = AutoTokenizer.from_pretrained("{repo_id}")

# Model is ready to use - no steering hooks needed!
```

## Citation

If you use this model, please cite Wisent:

```bibtex
@software{{wisent,
  title = {{Wisent: Contrastive Activation Addition}},
  author = {{...}},
  year = {{2025}},
  url = {{https://github.com/wisent}}
}}
```

## License

This model inherits the license from the base model: {base_model}
"""

    return card




"""Sparse Autoencoder model and training loop."""

import torch
import numpy as np
from typing import Tuple
from wisent.core.utils.config_tools.constants import NORM_EPS, L1_DEFAULT_COEF, SAE_L1_COEF_DEFAULT, SAE_N_EPOCHS_DEFAULT, SAE_BATCH_SIZE_DEFAULT, SAE_HIDDEN_DIM_MULTIPLIER


# =============================================================================
# SPARSE AUTOENCODER FOR FEATURE ANALYSIS
# =============================================================================

class SparseAutoencoder(torch.nn.Module):
    """
    Sparse Autoencoder for finding interpretable features in activations.
    
    Architecture: input -> encoder (with ReLU) -> sparse features -> decoder -> reconstruction
    
    The encoder learns an overcomplete basis where each feature ideally represents
    one interpretable concept. Sparsity is enforced via L1 penalty.
    """
    
    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,  # Usually 4x-8x input_dim for overcomplete
        l1_coef: float = L1_DEFAULT_COEF,
        tied_weights: bool = True,
    ):
        super().__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.l1_coef = l1_coef
        self.tied_weights = tied_weights
        
        # Encoder: input -> hidden (sparse features)
        self.encoder = torch.nn.Linear(input_dim, hidden_dim, bias=True)
        
        # Decoder: hidden -> input (reconstruction)
        if tied_weights:
            # Tied weights: decoder weight = encoder weight transposed
            self.decoder_bias = torch.nn.Parameter(torch.zeros(input_dim))
        else:
            self.decoder = torch.nn.Linear(hidden_dim, input_dim, bias=True)
        
        # Initialize weights
        self._init_weights()
    
    def _init_weights(self):
        # Kaiming initialization for encoder
        torch.nn.init.kaiming_uniform_(self.encoder.weight, nonlinearity='relu')
        torch.nn.init.zeros_(self.encoder.bias)
        
        if not self.tied_weights:
            torch.nn.init.kaiming_uniform_(self.decoder.weight, nonlinearity='linear')
            torch.nn.init.zeros_(self.decoder.bias)
    
    def encode(self, x: torch.Tensor) -> torch.Tensor:
        """Encode input to sparse feature activations."""
        return torch.relu(self.encoder(x))
    
    def decode(self, features: torch.Tensor) -> torch.Tensor:
        """Decode sparse features back to input space."""
        if self.tied_weights:
            return torch.mm(features, self.encoder.weight) + self.decoder_bias
        else:
            return self.decoder(features)
    
    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Forward pass returning (reconstruction, sparse_features)."""
        features = self.encode(x)
        reconstruction = self.decode(features)
        return reconstruction, features
    
    def loss(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Compute total loss = reconstruction_loss + l1_coef * sparsity_loss."""
        reconstruction, features = self.forward(x)
        
        # MSE reconstruction loss
        recon_loss = torch.nn.functional.mse_loss(reconstruction, x)
        
        # L1 sparsity penalty on feature activations
        sparsity_loss = features.abs().mean()
        
        total_loss = recon_loss + self.l1_coef * sparsity_loss
        
        return total_loss, recon_loss, sparsity_loss


def train_sparse_autoencoder(
    activations: torch.Tensor,
    hidden_dim: int = None,
    l1_coef: float = SAE_L1_COEF_DEFAULT,
    n_epochs: int = SAE_N_EPOCHS_DEFAULT,
    batch_size: int = SAE_BATCH_SIZE_DEFAULT,
    lr: float = L1_DEFAULT_COEF,
    device: str = "cpu",
    verbose: bool = True,
) -> SparseAutoencoder:
    """
    Train a sparse autoencoder on activations.
    
    Args:
        activations: Tensor of shape (n_samples, input_dim)
        hidden_dim: Number of features to learn (default: 4x input_dim)
        l1_coef: L1 sparsity coefficient
        n_epochs: Number of training epochs
        batch_size: Batch size for training
        lr: Learning rate
        device: Device to train on
        verbose: Print training progress
    
    Returns:
        Trained SparseAutoencoder
    """
    input_dim = activations.shape[1]
    if hidden_dim is None:
        hidden_dim = input_dim * SAE_HIDDEN_DIM_MULTIPLIER  # overcomplete
    
    # Move data to device
    activations = activations.to(device).float()
    
    # Normalize activations (important for SAE training)
    mean = activations.mean(dim=0, keepdim=True)
    std = activations.std(dim=0, keepdim=True) + NORM_EPS
    activations_norm = (activations - mean) / std
    
    # Create SAE
    sae = SparseAutoencoder(
        input_dim=input_dim,
        hidden_dim=hidden_dim,
        l1_coef=l1_coef,
    ).to(device)
    
    # Optimizer
    optimizer = torch.optim.Adam(sae.parameters(), lr=lr)
    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=n_epochs)
    
    # Training loop
    n_samples = activations_norm.shape[0]
    
    for epoch in range(n_epochs):
        # Shuffle data
        perm = torch.randperm(n_samples)
        epoch_loss = 0.0
        epoch_recon = 0.0
        epoch_sparsity = 0.0
        n_batches = 0
        
        for i in range(0, n_samples, batch_size):
            batch_idx = perm[i:i+batch_size]
            batch = activations_norm[batch_idx]
            
            optimizer.zero_grad()
            loss, recon_loss, sparsity_loss = sae.loss(batch)
            loss.backward()
            optimizer.step()
            
            epoch_loss += loss.item()
            epoch_recon += recon_loss.item()
            epoch_sparsity += sparsity_loss.item()
            n_batches += 1
        
        scheduler.step()
        
        if verbose and (epoch + 1) % 100 == 0:
            avg_loss = epoch_loss / n_batches
            avg_recon = epoch_recon / n_batches
            avg_sparse = epoch_sparsity / n_batches
            
            # Compute sparsity stats
            with torch.no_grad():
                features = sae.encode(activations_norm)
                active_frac = (features > 0).float().mean().item()
                avg_active = (features > 0).sum(dim=1).float().mean().item()
            
            print(f"  Epoch {epoch+1}/{n_epochs}: loss={avg_loss:.4f} "
                  f"(recon={avg_recon:.4f}, sparse={avg_sparse:.4f}), "
                  f"active_features={avg_active:.1f}/{hidden_dim} ({active_frac*100:.1f}%)")
    
    # Store normalization parameters
    sae.mean = mean
    sae.std = std
    
    return sae



"""Extracted NLP evaluator helper methods: NLI cross-encoder and embedding similarity."""

from __future__ import annotations


class NLPEvaluatorHelpersMixin:
    """Mixin providing NLI and embedding helper methods for NLPEvaluator."""

    CE_MODEL_NAME = "cross-encoder/nli-deberta-v3-small"
    EMB_MODEL_NAME = "sentence-transformers/all-MiniLM-L6-v2"

    def _load_ce(self):
        """Load the NLI cross-encoder model.
        Cross-encoder models are small and load quickly. They run on CPU
        reasonably well. They provide strong performance for entailment tasks.
        """
        from sentence_transformers import CrossEncoder
        _CE = CrossEncoder(self.CE_MODEL_NAME)
        return _CE

    def _nli_pick_between(
        self, response: str, options: list[str],
    ) -> tuple[int | None, list[float], float]:
        """
        Compare entailment(response -> 'The correct option is: <opt_i>')
        for i in {A,B}.
        Returns: (pred_idx, [entA, entB], margin)
        """
        ce = self._load_ce()
        pairs = [(response, f"The correct option is: {opt}") for opt in options]
        import torch
        import torch.nn.functional as F
        logits = torch.tensor(ce.predict(pairs))
        probs = F.softmax(logits, dim=-1).tolist()
        ent = [p[1] for p in probs]
        pred_idx = 1 if ent[0] > ent[1] else 2
        margin = abs(ent[0] - ent[1])
        return pred_idx, ent, margin

    def _nli_entailment_pair(
        self, a: str, bnormalize_text: str,
    ) -> tuple[float | None, float | None]:
        """Entailment probabilities for (a -> b) and (b -> a)."""
        try:
            ce = self._load_ce()
        except Exception:
            return None, None
        pairs = [(a, bnormalize_text), (bnormalize_text, a)]
        import torch
        import torch.nn.functional as F
        logits = torch.tensor(ce.predict(pairs))
        probs = F.softmax(logits, dim=-1).tolist()
        return probs[0][1], probs[1][1]

    def _load_emb(self):
        """Load the sentence embedding model."""
        from sentence_transformers import SentenceTransformer
        _EMB = SentenceTransformer(self.EMB_MODEL_NAME)
        return _EMB

    def _emb_sim(self, a: str, b: str) -> float | None:
        """Compute cosine similarity between two texts."""
        try:
            emb = self._load_emb()
        except Exception:
            return None
        import torch
        va, vb = emb.encode(
            [a, b], convert_to_tensor=True, normalize_embeddings=True)
        return torch.matmul(va, vb).item()

    def _emb_sims(
        self, response: str, options: list[str],
    ) -> tuple[float | None, float | None]:
        """Compute cosine similarities between response and two options."""
        try:
            emb = self._load_emb()
        except Exception:
            return None, None
        import torch
        vecs = emb.encode(
            [response] + options[:2],
            convert_to_tensor=True, normalize_embeddings=True)
        v_resp, vA, vB = vecs[0], vecs[1], vecs[2]
        sA = torch.matmul(v_resp, vA).item()
        sB = torch.matmul(v_resp, vB).item()
        return sA, sB



"""Helper mixins for FullBenchmarkDownloader split across multiple files."""

from ._converters_basic import BasicConvertersMixin
from ._converters_text import TextConvertersMixin
from ._converters_advanced import AdvancedConvertersMixin
from ._converters_qa import QAConvertersMixin

__all__ = [
    "BasicConvertersMixin",
    "TextConvertersMixin",
    "AdvancedConvertersMixin",
    "QAConvertersMixin",
]



"""Evaluate responses command execution logic."""

import json
import os
from pathlib import Path

from wisent.core.utils.config_tools.constants import JSON_INDENT
from wisent.core.utils.infra_tools.errors import TaskNotFoundError
from wisent.core.reading.evaluators.steering_evaluators import (
    SteeringEvaluatorFactory,
    EvaluatorConfig,
    RefusalEvaluator,
    PersonalizationEvaluator as SteeringPersonalizationEvaluator,
)
from wisent.core.reading.evaluators.rotator import EvaluatorRotator


from wisent.core.utils.cli.analysis.evaluation.evaluate_docker import evaluate_docker_execution
from wisent.core.utils.cli.analysis.evaluation.evaluate_personalization import evaluate_personalization
from wisent.core.utils.cli.analysis.evaluation.evaluate_refusal import evaluate_refusal
from wisent.core.utils.cli.analysis.evaluation.evaluate_standard import evaluate_standard


def execute_evaluate_responses(args):
    """
    Execute the evaluate-responses command.

    Evaluates generated responses using benchmark-specific evaluators.
    Uses EvaluatorRotator to auto-select the appropriate evaluator based on task name.

    Uses unified split strategy: all available splits are combined and split 80/20.
    Evaluation uses the TEST portion (20%) to ensure no data leakage with training.
    """
    from lm_eval.tasks import TaskManager
    from wisent.core.reading.evaluators.benchmark_specific import (
        GenerationEvaluator,
        ExactMatchEvaluator,
        F1Evaluator,
        PerplexityEvaluator
    )
    from wisent.core.utils import get_all_docs_from_task, create_deterministic_split

    print(f"\n{'='*80}")
    print(f"📊 EVALUATING GENERATED RESPONSES")
    print(f"{'='*80}")
    print(f"   Input: {args.input}")
    print(f"{'='*80}\n")

    # Load input file
    print(f"📂 Loading generated responses...")
    try:
        with open(args.input, 'r') as f:
            input_data = json.load(f)

        # Handle both dict format (with 'responses' key) and direct list format
        if isinstance(input_data, list):
            responses = input_data
            task_name = args.task if args.task else "generic"  # Default to generic for list format
        else:
            responses = input_data.get('responses', [])
            task_name = args.task if args.task else input_data.get('task', "generic")

        if not task_name:
            raise ValueError("Task name not found in input file and not provided via --task")
        print(f"   ✓ Loaded {len(responses)} responses")
        print(f"   Task: {task_name}\n")
    except Exception as e:
        print(f"   ❌ Failed to load input file: {e}")
        raise

    # Load task-evaluator mapping
    print(f"📋 Loading task evaluation config...")
    try:
        # Find task-evaluator.json in project root
        current_dir = Path(__file__).resolve()
        project_root = None
        for parent in current_dir.parents:
            task_eval_file = parent / 'task-evaluator.json'
            if task_eval_file.exists():
                project_root = parent
                break

        if not project_root:
            import wisent
            wisent_pkg_dir = Path(wisent.__file__).resolve().parent
            pkg_task_eval = wisent_pkg_dir / "task-evaluator.json"
            if pkg_task_eval.exists():
                project_root = wisent_pkg_dir

        if not project_root:
            raise FileNotFoundError("Could not find task-evaluator.json")
        with open(project_root / 'task-evaluator.json', 'r') as f:
            task_evaluator_map = json.load(f)

        # Get task evaluation type
        task_config = task_evaluator_map['tasks'].get(task_name)
        if not task_config:
            raise TaskNotFoundError(task_name=task_name)

        evaluation_type = task_config['evaluation_type']
        primary_metric = task_config['primary_metric']

        print(f"   ✓ Task evaluation type: {evaluation_type}")
        print(f"   ✓ Primary metric: {primary_metric}\n")
    except Exception as e:
        print(f"   ❌ Could not load task config: {e}")
        raise

    # Load task to get ground truth (skip for docker_execution, personalization, or if responses have references)
    task_docs = None
    task = None
    
    # Check if all responses already have positive_reference (ground truth from generation)
    has_references = all(
        r.get('positive_reference') is not None 
        for r in responses
    )
    
    if evaluation_type not in ["docker_execution", "personalization"] and not has_references:
        print(f"📚 Loading task data using unified split strategy...")
        try:
            tm = TaskManager()
            task_dict = tm.load_task_or_group(task_name)
            task = task_dict[task_name]

            # Use unified split strategy: combine ALL available splits, then use TEST portion (20%)
            all_docs, split_counts = get_all_docs_from_task(task)
            _, task_docs = create_deterministic_split(all_docs, task_name)

            print(f"   ✓ Combined {len(all_docs)} total docs from splits: {split_counts}")
            print(f"   ✓ Using TEST portion: {len(task_docs)} task documents for evaluation\n")
        except Exception as e:
            print(f"   ❌ Could not load task: {e}")
            raise
    elif has_references:
        print(f"📚 Using references from responses file (task loading skipped)\n")

    # Select evaluator using EvaluatorRotator (auto-selects based on task name)
    print(f"🔧 Selecting evaluator for task '{task_name}'...")
    evaluator = None
    evaluator_rotator = None
    
    # Special handling for certain evaluation types
    if evaluation_type == "docker_execution":
        from wisent.core.reading.evaluators.benchmark_specific.coding.metrics.evaluator import (
            CodingEvaluator,
            EvaluatorConfig
        )
        from wisent.core.reading.evaluators.benchmark_specific.coding.providers.livecodebench import LiveCodeBenchProvider
        print(f"   Using: CodingEvaluator (Docker sandbox execution)\n")
        docker_config = task_config.get('docker_config', {})
        provider_name = task_config.get('provider', 'livecodebench')
    elif evaluation_type == "personalization":
        print(f"   Using: SteeringPersonalizationEvaluator\n")
    elif evaluation_type == "refusal":
        print(f"   Using: RefusalEvaluator\n")
    else:
        # Use EvaluatorRotator to auto-select based on task name
        try:
            EvaluatorRotator.discover_evaluators("wisent.core.reading.evaluators.oracles")
            EvaluatorRotator.discover_evaluators("wisent.core.reading.evaluators.benchmark_specific")
            evaluator_rotator = EvaluatorRotator(evaluator=None, task_name=task_name)
            evaluator = evaluator_rotator.current
            print(f"   Using: {evaluator.name} (auto-selected via EvaluatorRotator)\n")
        except Exception as e:
            # Fallback to manual selection if rotator fails
            print(f"   ⚠️  EvaluatorRotator failed: {e}")
            print(f"   Falling back to manual selection based on evaluation_type...")
            if evaluation_type == "multiple_choice":
                evaluator = F1Evaluator()
                print(f"   Using: F1Evaluator (compares response to choice texts)\n")
            elif evaluation_type == "generate_until":
                if primary_metric == "exact_match":
                    evaluator = ExactMatchEvaluator()
                    print(f"   Using: ExactMatchEvaluator (extracts and compares answers)\n")
                elif primary_metric in ["em", "f1"]:
                    evaluator = F1Evaluator()
                    print(f"   Using: F1Evaluator (token-level comparison)\n")
                else:
                    evaluator = GenerationEvaluator()
                    print(f"   Using: GenerationEvaluator (extracts and compares answers)\n")
            elif evaluation_type == "loglikelihood_rolling":
                evaluator = PerplexityEvaluator()
                print(f"   Using: PerplexityEvaluator (perplexity computation)\n")
            else:
                evaluator = F1Evaluator()
                print(f"   Using: F1Evaluator (default fallback)\n")

    # Evaluate responses
    print(f"🎯 Evaluating responses...\n")
    evaluation_results = []
    task_results = []  # For aggregation


    # Handle docker_execution separately
    if evaluation_type == "docker_execution":
        aggregated_metrics = evaluate_docker_execution(
            args, input_data, responses, task_name, evaluation_results, task_results,
            task_config=task_config)
        if aggregated_metrics is not None:
            return

    # Handle personalization separately
    if evaluation_type == "personalization":
        aggregated_metrics = evaluate_personalization(
            args, input_data, responses, task_name, evaluation_results, task_results)
        if aggregated_metrics is not None:
            return

    # Handle refusal evaluation
    if evaluation_type == "refusal":
        aggregated_metrics = evaluate_refusal(
            args, input_data, responses, task_name, evaluation_results, task_results)
        if aggregated_metrics is not None:
            return

    # Standard evaluation loop
    evaluate_standard(
        args, input_data, responses, task_name,
        evaluation_results, task_results, evaluator, task_docs,
        evaluation_type=evaluation_type,
    )

    # Aggregate results
    aggregated_metrics = {}
    if task_results:
        # Get all metric keys
        all_metric_keys = set()
        for result in task_results:
            all_metric_keys.update(result.keys())

        # Aggregate each metric
        for metric_key in all_metric_keys:
            values = [r[metric_key] for r in task_results if metric_key in r]
            if values:
                # Most tasks use mean aggregation
                aggregated_metrics[metric_key] = sum(values) / len(values)

    # Save results
    print(f"\n💾 Saving evaluation results...")
    os.makedirs(os.path.dirname(args.output), exist_ok=True)

    output_data = {
        "input_file": args.input,
        "task": task_name if isinstance(input_data, list) else input_data.get('task'),
        "model": None if isinstance(input_data, list) else input_data.get('model'),
        "evaluation_type": evaluation_type,
        "evaluator_used": evaluator.name,
        "aggregated_metrics": aggregated_metrics,
        "num_evaluated": len(task_results),
        "num_total": len(responses),
        "evaluations": evaluation_results
    }

    with open(args.output, 'w') as f:
        json.dump(output_data, f, indent=JSON_INDENT)

    print(f"   ✓ Results saved to: {args.output}\n")

    # Print summary
    print(f"{'='*80}")
    print(f"✅ EVALUATION COMPLETE")
    print(f"{'='*80}")
    print(f"   Evaluator: {evaluator.name}")
    print(f"   Total responses: {len(responses)}")
    print(f"   Successfully evaluated: {len(task_results)}")
    print(f"   Failed/skipped: {len(responses) - len(task_results)}")
    print(f"\n   Metrics:")
    for metric_name, metric_value in aggregated_metrics.items():
        if isinstance(metric_value, float):
            print(f"     {metric_name}: {metric_value:.4f}")
        else:
            print(f"     {metric_name}: {metric_value}")
    print(f"{'='*80}\n")



"""Task manager package for lm-evaluation-harness integration."""

from .core import (
    load_available_tasks,
    load_docs,
    get_available_tasks,
    is_valid_task,
    resolve_task_name,
    TaskManager,
    _task_manager,
)

from .group_handling import (
    find_working_task_from_group,
    handle_configurable_group_task,
)

from .extraction import (
    extract_individual_tasks_from_yaml,
    try_find_related_working_task,
    try_extract_working_tasks_from_group,
)

from .yaml_support import (
    save_custom_task_yaml,
    create_task_yaml_from_user_content,
    load_with_env_config,
    create_flan_held_in_files,
    load_task_with_config_dir,
)

__all__ = [
    # Core functions
    'load_available_tasks',
    'load_docs',
    'get_available_tasks',
    'is_valid_task',
    'resolve_task_name',
    'TaskManager',
    '_task_manager',
    # Group handling
    'find_working_task_from_group',
    'handle_configurable_group_task',
    # Extraction
    'extract_individual_tasks_from_yaml',
    'try_find_related_working_task',
    'try_extract_working_tasks_from_group',
    # YAML support
    'save_custom_task_yaml',
    'create_task_yaml_from_user_content',
    'load_with_env_config',
    'create_flan_held_in_files',
    'load_task_with_config_dir',
]



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from wisent.core.primitives.contrastive_pairs.core.pair import ContrastivePair
from wisent.core.primitives.contrastive_pairs.core.io.response import NegativeResponse, PositiveResponse
from wisent.extractors.lm_eval.atoms import LMEvalBenchmarkExtractor
from wisent.core.utils.cli.cli_logger import setup_logger, bind
from wisent.core.utils.config_tools.constants import (
    PALOMA_CHUNK_SIZE_WORDS,
    PALOMA_DOCS_DIVISOR,
    PALOMA_MAX_CHARS,
    PALOMA_MAX_WORDS,
    PALOMA_MIN_CHUNK_WORDS,
    PALOMA_MIN_TEXT_LENGTH,
    PALOMA_MIN_WORDS,
    PALOMA_TARGET_PAIRS_PER_DOC,
)

if TYPE_CHECKING:
    from lm_eval.api.task import ConfigurableTask


__all__ = ["PalomaExtractor"]
_LOG = setup_logger(__name__)

task_names = (
    "paloma_4chan_meta_sep", "paloma_c4_100_domains", "paloma_c4_en",
    "paloma_dolma_100_programing_languages", "paloma_dolma_100_subreddits",
    "paloma_dolma-v1_5", "paloma_falcon-refinedweb", "paloma_gab",
    "paloma_m2d2_s2orc_unsplit", "paloma_m2d2_wikipedia_unsplit",
    "paloma_manosphere_meta_sep", "paloma_mc4", "paloma_ptb",
    "paloma_redpajama", "paloma_twitterAAE_HELM_fixed", "paloma_wikitext_103"
)

# Note: Paloma is a perplexity benchmark (loglikelihood_rolling)
# We use perplexity evaluation with text corruption for contrastive pairs
class PalomaExtractor(LMEvalBenchmarkExtractor):
    """Extractor for the Paloma benchmark."""


    evaluator_name = "log_likelihoods"
    def extract_contrastive_pairs(
        self,
        lm_eval_task_data: ConfigurableTask,
        limit: int | None = None,
        preferred_doc: str | None = None,
    ) -> list[ContrastivePair]:
        """
        Build contrastive pairs from Paloma docs.

        Note: Paloma docs are very long (hundreds of thousands of characters).
        We split each doc into multiple chunks to generate multiple pairs.

        Args:
            lm_eval_task_data: lm-eval task instance for Paloma.
            limit: Optional maximum number of pairs to produce.
            preferred_doc: Optional preferred document source.

        Returns:
            A list of ContrastivePair objects.
        """
        log = bind(_LOG, task=getattr(lm_eval_task_data, "NAME", "unknown"))

        max_items = self._normalize_limit(limit)
        # Load fewer docs since we'll split each into multiple pairs
        docs_to_load = 1 if max_items is None else max(1, max_items // PALOMA_DOCS_DIVISOR)
        docs = self.load_docs(lm_eval_task_data, docs_to_load, preferred_doc=preferred_doc)

        pairs: list[ContrastivePair] = []

        log.info("Extracting contrastive pairs", extra={"doc_count": len(docs)})

        for doc in docs:
            # Extract multiple pairs from each long document
            doc_pairs = self._extract_pairs_from_doc(doc, max_pairs=max_items - len(pairs) if max_items else None)
            pairs.extend(doc_pairs)
            if max_items is not None and len(pairs) >= max_items:
                pairs = pairs[:max_items]
                break

        if not pairs:
            task_name = getattr(lm_eval_task_data, "NAME", type(lm_eval_task_data).__name__)
            log.warning("No valid Paloma pairs extracted", extra={"task": task_name})

        return pairs

    def _extract_pairs_from_doc(self, doc: dict[str, Any], max_pairs: int | None = None) -> list[ContrastivePair]:
        """
        Extract multiple contrastive pairs from a single long Paloma document.

        Args:
            doc: Paloma document with 'text' field
            max_pairs: Maximum number of pairs to extract from this doc

        Returns:
            List of ContrastivePair objects
        """
        pairs = []
        text = doc.get("text", "").strip()

        if not text or len(text) < PALOMA_MIN_TEXT_LENGTH:
            return pairs

        # Split long text into chunks of ~PALOMA_CHUNK_SIZE_WORDS words
        words = text.split()
        chunk_size = PALOMA_CHUNK_SIZE_WORDS
        target_pairs = max_pairs if max_pairs else PALOMA_TARGET_PAIRS_PER_DOC

        for i in range(target_pairs):
            start_idx = i * chunk_size
            if start_idx >= len(words):
                break

            end_idx = min(start_idx + chunk_size, len(words))
            chunk_words = words[start_idx:end_idx]

            if len(chunk_words) < PALOMA_MIN_CHUNK_WORDS:  # Skip very short chunks
                continue

            chunk_text = " ".join(chunk_words)

            # Create corrupted version
            mid_start = len(chunk_words) // 3
            mid_end = 2 * len(chunk_words) // 3
            middle_words = chunk_words[mid_start:mid_end]

            import random
            random.seed(hash(chunk_text) % (2**32))
            shuffled_middle = middle_words.copy()
            random.shuffle(shuffled_middle)

            corrupted_words = chunk_words[:mid_start] + shuffled_middle + chunk_words[mid_end:]
            corrupted_text = " ".join(corrupted_words)

            metadata = {"label": "paloma", "task": doc.get("source", "paloma"), "chunk": i}

            pair = self._build_pair(
                question="Continue this text:",
                correct=chunk_text,
                incorrect=corrupted_text,
                metadata=metadata,
            )
            pairs.append(pair)

        return pairs

    def _extract_pair_from_doc(self, doc: dict[str, Any]) -> ContrastivePair | None:
        """
        Convert a single Paloma doc into a ContrastivePair for perplexity evaluation.

        Paloma is a perplexity benchmark with only text passages. We create contrastive
        pairs by using the original text as positive and a corrupted version as negative.
        The perplexity evaluator will compare both and should prefer the original.
        """
        log = bind(_LOG, doc_id=doc.get("id", "unknown"))

        try:
            # Paloma docs have a 'text' field with the passage
            text = doc.get("text", "").strip()

            if not text or len(text) < PALOMA_MIN_CHUNK_WORDS:  # Skip very short texts
                log.debug("Skipping doc with insufficient text", extra={"text_len": len(text)})
                return None

            # Take a reasonable chunk (first ~PALOMA_MAX_WORDS words or ~PALOMA_MAX_CHARS chars)
            words = text.split()
            if len(words) > PALOMA_MAX_WORDS:
                text = " ".join(words[:PALOMA_MAX_WORDS])
            elif len(text) > PALOMA_MAX_CHARS:
                text = text[:PALOMA_MAX_CHARS]

            # Create corrupted version by shuffling some words
            # This should have higher perplexity than the original
            words = text.split()
            if len(words) < PALOMA_MIN_WORDS:
                return None

            # Shuffle middle 30% of words to create unnatural text
            mid_start = len(words) // 3
            mid_end = 2 * len(words) // 3
            middle_words = words[mid_start:mid_end]

            import random
            random.seed(hash(text) % (2**32))  # Deterministic shuffle based on text
            shuffled_middle = middle_words.copy()
            random.shuffle(shuffled_middle)

            corrupted_words = words[:mid_start] + shuffled_middle + words[mid_end:]
            corrupted_text = " ".join(corrupted_words)

            # Build the pair with minimal prompt (perplexity is computed on the response)
            metadata = {"label": "paloma", "task": doc.get("source", "paloma")}

            return self._build_pair(
                question="Continue this text:",  # Minimal prompt for perplexity tasks
                correct=text,
                incorrect=corrupted_text,
                metadata=metadata,
            )

        except Exception as exc:
            log.error("Error extracting pair from doc", exc_info=exc, extra={"doc": doc})
            return None

    @staticmethod
    def _build_pair(
        question: str,
        correct: str,
        incorrect: str,
        metadata: dict[str, Any] | None = None,
    ) -> ContrastivePair:
        positive_response = PositiveResponse(model_response=correct)
        negative_response = NegativeResponse(model_response=incorrect)
        return ContrastivePair(prompt=question, positive_response=positive_response, negative_response=negative_response, label=metadata.get("label"))



"""Activation analysis helpers for verify-steering."""
import torch
from typing import Any, Dict, List, Optional
from wisent.core.utils.config_tools.constants import (
    COMPARE_TOL,
    DISPLAY_TRUNCATION_SHORT,
    MAX_NEW_TOKENS_VERIFY_SINGLE,
    SEPARATOR_WIDTH_MEDIUM_PLUS,
    SEPARATOR_WIDTH_STANDARD,
    SEPARATOR_WIDTH_WIDE,
    VERIFY_ALIGNMENT_THRESHOLD,
    VERIFY_GATE_ACTIVE_THRESHOLD,
    VERIFY_GATE_PARTIAL_THRESHOLD,
    VERIFY_GATE_VARIANCE_THRESHOLD,
    VERIFY_GOOD_LAYER_THRESHOLD,
)


def _compare_activations(
    base_model,
    steered_model,
    tokenizer,
    prompts: List[str],
    steering_dirs: Dict[str, torch.Tensor],
    verbose: bool = False,
) -> Dict[str, Any]:
    """Compare activations between base and steered models."""
    # Get layer modules
    if hasattr(base_model, "model"):
        base_layers = base_model.model.layers
        steered_layers = steered_model.model.layers
    elif hasattr(base_model, "transformer"):
        base_layers = base_model.transformer.h
        steered_layers = steered_model.transformer.h
    else:
        base_layers = getattr(base_model, "layers", [])
        steered_layers = getattr(steered_model, "layers", [])

    # Map layer names to indices
    layer_name_to_idx = {}
    for layer_name in steering_dirs.keys():
        try:
            idx = int(str(layer_name).split("_")[-1]) if "_" in str(layer_name) else int(layer_name)
            layer_name_to_idx[layer_name] = idx
        except (ValueError, IndexError):
            pass

    all_results = []

    for prompt in prompts:
        # Format prompt
        messages = [{"role": "user", "content": prompt}]
        try:
            formatted = tokenizer.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)
        except Exception:
            formatted = prompt

        inputs = tokenizer(formatted, return_tensors="pt")
        inputs_base = {k: v.to(base_model.device) for k, v in inputs.items()}
        inputs_steered = {k: v.to(steered_model.device) for k, v in inputs.items()}

        # Capture activations
        base_acts = {}
        steered_acts = {}

        def make_hook(storage, layer_name):
            def hook(module, input, output):
                hidden = output[0] if isinstance(output, tuple) else output
                storage[layer_name] = hidden[:, -1, :].detach().clone()
                return output
            return hook

        # Install hooks
        hooks = []
        for layer_name, layer_idx in layer_name_to_idx.items():
            if layer_idx < len(base_layers):
                h1 = base_layers[layer_idx].register_forward_hook(make_hook(base_acts, layer_name))
                h2 = steered_layers[layer_idx].register_forward_hook(make_hook(steered_acts, layer_name))
                hooks.extend([h1, h2])

        # Forward pass
        with torch.no_grad():
            _ = base_model(**inputs_base)
            _ = steered_model(**inputs_steered)

        # Remove hooks
        for h in hooks:
            h.remove()

        # Compute alignments
        prompt_result = {"prompt": prompt[:DISPLAY_TRUNCATION_SHORT], "layers": {}}

        for layer_name, direction in steering_dirs.items():
            if layer_name in base_acts and layer_name in steered_acts:
                base_act = base_acts[layer_name][0].float()
                steered_act = steered_acts[layer_name][0].float().to(base_act.device)
                direction = direction.to(base_act.device).float()

                shift = steered_act - base_act
                shift_norm = shift.norm().item()

                if shift_norm > COMPARE_TOL:
                    alignment = F.cosine_similarity(
                        shift.unsqueeze(0), direction.unsqueeze(0)
                    ).item()
                else:
                    alignment = 0.0

                prompt_result["layers"][layer_name] = {
                    "shift_norm": shift_norm,
                    "alignment": alignment,
                    "base_norm": base_act.norm().item(),
                    "steered_norm": steered_act.norm().item(),
                }

        all_results.append(prompt_result)

    # Compute overall metrics
    all_alignments = []
    layer_alignments = {}

    for result in all_results:
        for layer_name, layer_data in result["layers"].items():
            all_alignments.append(layer_data["alignment"])
            if layer_name not in layer_alignments:
                layer_alignments[layer_name] = []
            layer_alignments[layer_name].append(layer_data["alignment"])

    overall_alignment = sum(all_alignments) / len(all_alignments) if all_alignments else 0

    # Print results
    print(f"\n{'Prompt':<40} {'Avg Alignment':<15} {'Status'}")
    print("-" * SEPARATOR_WIDTH_WIDE)

    for result in all_results:
        alignments = [v["alignment"] for v in result["layers"].values()]
        avg = sum(alignments) / len(alignments) if alignments else 0
        status = "ALIGNED" if avg > VERIFY_ALIGNMENT_THRESHOLD else ("WEAK" if avg > 0 else "MISALIGNED")
        print(f"{result['prompt']:<40} {avg:>12.4f}    {status}")

    if verbose:
        print("\n" + "-" * SEPARATOR_WIDTH_WIDE)
        print("Per-Layer Alignment:")
        print(f"{'Layer':<10} {'Avg Alignment':<15} {'Min':<10} {'Max':<10} {'Status'}")
        print("-" * SEPARATOR_WIDTH_MEDIUM_PLUS)

        for layer_name in sorted(layer_alignments.keys(), key=lambda x: int(x) if x.isdigit() else 0):
            aligns = layer_alignments[layer_name]
            avg = sum(aligns) / len(aligns)
            min_a = min(aligns)
            max_a = max(aligns)
            status = "OK" if avg > VERIFY_ALIGNMENT_THRESHOLD else ("WEAK" if avg > 0 else "BAD")
            print(f"{layer_name:<10} {avg:>12.4f}    {min_a:>8.4f}  {max_a:>8.4f}  {status}")

    return {
        "per_prompt": all_results,
        "per_layer": {k: sum(v) / len(v) for k, v in layer_alignments.items()},
        "overall_alignment": overall_alignment,
    }


def _check_gate_intensity(
    steered_model,
    tokenizer,
    steering_data: Dict,
    prompts: List[str],
    check_gate: bool,
    check_intensity: bool,
) -> tuple:
    """Check gate and intensity network predictions."""
    from wisent.core.weight_modification.standalone_loader import GROMHooks

    hooks = GROMHooks(steered_model, steering_data)
    hooks.install()

    gate_results = []
    intensity_results = []

    for prompt in prompts:
        messages = [{"role": "user", "content": prompt}]
        try:
            formatted = tokenizer.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)
        except Exception:
            formatted = prompt

        inputs = tokenizer(formatted, return_tensors="pt").to(steered_model.device)

        with torch.no_grad():
            _ = steered_model.generate(
                **inputs,
                max_new_tokens=MAX_NEW_TOKENS_VERIFY_SINGLE,
                do_sample=False,
                pad_token_id=tokenizer.eos_token_id,
            )

        if check_gate:
            gate = hooks.get_current_gate()
            gate_results.append({"prompt": prompt[:DISPLAY_TRUNCATION_SHORT], "gate": gate})

        if check_intensity:
            intensities = hooks.get_current_intensities()
            intensity_results.append({"prompt": prompt[:DISPLAY_TRUNCATION_SHORT], "intensities": intensities})

    hooks.remove()

    # Print gate results
    if gate_results:
        print(f"\nGate Values:")
        print(f"{'Prompt':<40} {'Gate':<10} {'Status'}")
        print("-" * SEPARATOR_WIDTH_STANDARD)
        for r in gate_results:
            gate = r["gate"]
            status = "ACTIVE" if gate > VERIFY_GATE_ACTIVE_THRESHOLD else ("PARTIAL" if gate > VERIFY_GATE_PARTIAL_THRESHOLD else "INACTIVE")
            print(f"{r['prompt']:<40} {gate:>8.4f}  {status}")

        # Check discrimination
        gates = [r["gate"] for r in gate_results]
        gate_std = torch.tensor(gates).std().item()
        if gate_std < VERIFY_GATE_VARIANCE_THRESHOLD:
            print(f"\n[WARNING] Gate not discriminating (std={gate_std:.4f})")
            print("  All prompts get similar gate values - gating network may need more training")

    # Print intensity summary
    if intensity_results:
        print(f"\nIntensity Summary:")
        for r in intensity_results:
            if r["intensities"]:
                avg = sum(r["intensities"].values()) / len(r["intensities"])
                print(f"  {r['prompt']}: avg={avg:.4f}")

    return gate_results, intensity_results


def _print_summary(
    results: Dict,
    gate_results: Optional[List],
    intensity_results: Optional[List],
    threshold: float,
):
    """Print verification summary."""
    overall = results.get("overall_alignment", 0)
    per_layer = results.get("per_layer", {})

    # Count layers by status
    good_layers = sum(1 for v in per_layer.values() if v > VERIFY_GOOD_LAYER_THRESHOLD)
    weak_layers = sum(1 for v in per_layer.values() if 0 < v <= VERIFY_GOOD_LAYER_THRESHOLD)
    bad_layers = sum(1 for v in per_layer.values() if v <= 0)

    print(f"\nOverall Alignment: {overall:.4f}")
    print(f"  - Threshold: {threshold}")
    print(f"  - Status: {'PASS' if overall >= threshold else 'FAIL'}")

    print(f"\nLayer Breakdown:")
    print(f"  - Well-aligned (>0.5): {good_layers}")
    print(f"  - Weak alignment (0-0.5): {weak_layers}")
    print(f"  - Misaligned (<0): {bad_layers}")

    if bad_layers > 0:
        print("\n[WARNING] Some layers have negative alignment!")
        print("  The steering direction is OPPOSITE to the activation shift.")
        print("  This will likely cause steering to fail or have opposite effect.")
        bad_layer_names = [k for k, v in per_layer.items() if v <= 0]
        print(f"  Problematic layers: {bad_layer_names}")

    if gate_results:
        gates = [r["gate"] for r in gate_results]
        avg_gate = sum(gates) / len(gates)
        print(f"\nGate Network:")
        print(f"  - Average gate: {avg_gate:.4f}")
        print(f"  - Std dev: {torch.tensor(gates).std().item():.4f}")

    print("\nInterpretation Guide:")
    print("  - Alignment > 0.5: Steering working correctly")
    print("  - Alignment 0-0.5: Steering is weak")
    print("  - Alignment < 0: Steering going WRONG direction")
    print("  - Gate ~0.5 for all: Gate not discriminating (needs more data)")



"""
Test for train_and_multi_steer.sh example.

This test validates training multiple steering vectors and combining them
with different weights during inference using the multi-steer command.
"""

import subprocess
import pytest
import tempfile
import os


def test_train_multiple_vectors_and_combine():
    """Test training multiple vectors and combining them with multi-steer."""
    with tempfile.TemporaryDirectory() as tmpdir:
        vector1_path = os.path.join(tmpdir, "vector1.pt")
        vector2_path = os.path.join(tmpdir, "vector2.pt")
        combined_path = os.path.join(tmpdir, "combined.pt")
        
        # Train first vector
        train1_result = subprocess.run(
            [
                "python", "-m", "wisent.core.primitives.model_interface.core.main", "generate-vector-from-task",
                "--task", "boolq",
                "--trait-label", "truthfulness",
                "--output", vector1_path,
                "--model", "meta-llama/Llama-3.2-1B-Instruct",
                "--num-pairs", "10",
                "--layers", "3",
                "--token-aggregation", "average",
                "--method", "caa",
                "--normalize",
                "--device", "cpu",
                "--verbose"
            ],
            capture_output=True,
            text=True,
            timeout=300
        )
        
        assert train1_result.returncode == 0, f"Training vector1 failed: {train1_result.stderr}"
        assert os.path.exists(vector1_path), "Vector1 was not saved"
        
        # Train second vector
        train2_result = subprocess.run(
            [
                "python", "-m", "wisent.core.primitives.model_interface.core.main", "generate-vector-from-task",
                "--task", "boolq",
                "--trait-label", "helpfulness",
                "--output", vector2_path,
                "--model", "meta-llama/Llama-3.2-1B-Instruct",
                "--num-pairs", "10",
                "--layers", "3",
                "--token-aggregation", "average",
                "--method", "caa",
                "--normalize",
                "--device", "cpu",
                "--verbose"
            ],
            capture_output=True,
            text=True,
            timeout=300
        )
        
        assert train2_result.returncode == 0, f"Training vector2 failed: {train2_result.stderr}"
        assert os.path.exists(vector2_path), "Vector2 was not saved"
        
        # Combine vectors with equal weights
        combine_result = subprocess.run(
            [
                "python", "-m", "wisent.core.primitives.model_interface.core.main", "multi-steer",
                "--vector", f"{vector1_path}:0.5",
                "--vector", f"{vector2_path}:0.5",
                "--model", "meta-llama/Llama-3.2-1B-Instruct",
                "--layer", "3",
                "--method", "CAA",
                "--prompt", "What is AI?",
                "--max-new-tokens", "50",
                "--normalize-weights",
                "--save-combined", combined_path,
                "--device", "cpu",
                "--verbose"
            ],
            capture_output=True,
            text=True,
            timeout=180
        )
        
        # Should complete without error
        assert combine_result.returncode == 0, f"Combining failed: {combine_result.stderr}"
        
        # Combined vector should be saved
        assert os.path.exists(combined_path), "Combined vector was not saved"
        
        # Output should contain generated text
        assert len(combine_result.stdout) > 0, "No output generated"


def test_multi_steer_with_different_emphasis():
    """Test combining vectors with different weight emphasis."""
    with tempfile.TemporaryDirectory() as tmpdir:
        vector1_path = os.path.join(tmpdir, "vector1.pt")
        vector2_path = os.path.join(tmpdir, "vector2.pt")
        
        # Train vectors (abbreviated for test)
        train1_result = subprocess.run(
            [
                "python", "-m", "wisent.core.primitives.model_interface.core.main", "generate-vector-from-task",
                "--task", "boolq",
                "--trait-label", "trait1",
                "--output", vector1_path,
                "--model", "meta-llama/Llama-3.2-1B-Instruct",
                "--num-pairs", "10",
                "--layers", "3",
                "--token-aggregation", "average",
                "--method", "caa",
                "--normalize",
                "--device", "cpu"
            ],
            capture_output=True,
            text=True,
            timeout=300
        )
        
        assert train1_result.returncode == 0
        
        train2_result = subprocess.run(
            [
                "python", "-m", "wisent.core.primitives.model_interface.core.main", "generate-vector-from-task",
                "--task", "boolq",
                "--trait-label", "trait2",
                "--output", vector2_path,
                "--model", "meta-llama/Llama-3.2-1B-Instruct",
                "--num-pairs", "10",
                "--layers", "3",
                "--token-aggregation", "average",
                "--method", "caa",
                "--normalize",
                "--device", "cpu"
            ],
            capture_output=True,
            text=True,
            timeout=300
        )
        
        assert train2_result.returncode == 0
        
        # Emphasize first vector (70% vs 30%)
        combine_result = subprocess.run(
            [
                "python", "-m", "wisent.core.primitives.model_interface.core.main", "multi-steer",
                "--vector", f"{vector1_path}:0.7",
                "--vector", f"{vector2_path}:0.3",
                "--model", "meta-llama/Llama-3.2-1B-Instruct",
                "--layer", "3",
                "--method", "CAA",
                "--prompt", "Explain quantum computing.",
                "--max-new-tokens", "50",
                "--normalize-weights",
                "--device", "cpu",
                "--verbose"
            ],
            capture_output=True,
            text=True,
            timeout=180
        )
        
        # Should complete without error
        assert combine_result.returncode == 0, f"Combining failed: {combine_result.stderr}"


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""Score group task manifest."""

from __future__ import annotations

BASE_IMPORT = "wisent.extractors.lm_eval.lm_task_extractors."

SCORE_TASKS = {
    "score_robustness": f"{BASE_IMPORT}score:ScoreExtractor",
    "score_robustness_agieval": f"{BASE_IMPORT}score:ScoreExtractor",
    "score_robustness_math": f"{BASE_IMPORT}score:ScoreExtractor",
    "score_robustness_mmlu_pro": f"{BASE_IMPORT}score:ScoreExtractor",
    "score_prompt_robustness_agieval": f"{BASE_IMPORT}score:ScoreExtractor",
    "score_prompt_robustness_math": f"{BASE_IMPORT}score:ScoreExtractor",
    "score_prompt_robustness_mmlu_pro": f"{BASE_IMPORT}score:ScoreExtractor",
    "score_option_order_robustness_agieval": f"{BASE_IMPORT}score:ScoreExtractor",
    "score_option_order_robustness_mmlu_pro": f"{BASE_IMPORT}score:ScoreExtractor",
    "score_non_greedy_robustness_agieval": f"{BASE_IMPORT}score:ScoreExtractor",
    "score_non_greedy_robustness_math": f"{BASE_IMPORT}score:ScoreExtractor",
    "score_non_greedy_robustness_mmlu_pro": f"{BASE_IMPORT}score:ScoreExtractor",
}



"""WisentModel methods: apply_steering_object, _encode_one, _batch_encode, _extract_assistant_response."""
from __future__ import annotations

import math
import re
from typing import Any

import torch
import torch.nn as nn

from wisent.core.primitives.models.layer import extract_token_ids
from wisent.core.primitives.models.core.atoms import SteeringVector
from wisent.core.control.generation.prompts.core.atom import ChatMessage
from wisent.core.utils.infra_tools.errors import ChatTemplateNotAvailableError
from wisent.core.utils.config_tools.constants import (
    DEFAULT_STRENGTH,
    STEERING_APP_DECAY_RATE,
    STEERING_APP_GAUSSIAN_CENTER,
    STEERING_APP_GAUSSIAN_WIDTH,
    STEERING_APP_INITIAL_TOKENS,
)
from wisent.core.primitives.models.config import get_generate_kwargs


def _apply_steering_object(
    self,
    steering_obj: "BaseSteeringObject",
    base_strength: float = DEFAULT_STRENGTH,
    steering_strategy: str = "constant",
    steering_strategy_config: dict | None = None,
    max_new_tokens: int | None = None,
) -> None:
    """
    Register forward hooks using a SteeringObject with full method-specific logic.

    Unlike apply_steering() which uses simple vector addition, this method
    uses the SteeringObject's compute_gate() and compute_intensity() methods
    for conditional steering (TETNO, GROM).

    Args:
        steering_obj: A SteeringObject (CAA, Ostrze, MLP, TECZA, TETNO, or GROM)
        base_strength: Base multiplier for steering intensity
        steering_strategy: How to apply steering over tokens:
            - "constant": Same strength throughout generation
            - "initial_only": Only apply at first N tokens
            - "diminishing": Strength decreases over tokens
            - "increasing": Strength increases over tokens
            - "gaussian": Gaussian curve centered at specific token position
        steering_strategy_config: Config for strategy (initial_tokens, rate, gaussian_center, gaussian_width)
        max_new_tokens: Expected max tokens for strategy weight calculation

    Example:
        >>> from wisent.core.control.steering_methods.steering_object import BaseSteeringObject
        >>> obj = BaseSteeringObject.load("steering.pt")
        >>> wm.apply_steering_object(obj, base_strength=1.0, steering_strategy="diminishing")
        >>> response = wm.generate([...])
        >>> wm.detach()
    """
    from wisent.core.control.steering_methods.steering_object import BaseSteeringObject

    if max_new_tokens is None:
        max_new_tokens = get_generate_kwargs()["max_new_tokens"]

    self.detach()
    if hasattr(steering_obj, 'set_model_weights'):
        steering_obj.set_model_weights(self.hf_model)
    config = steering_strategy_config or {}

    # Token counter shared across all hooks (mutable container)
    token_counter = {"count": 0, "prompt_len": 0}

    def get_steering_weight(strategy: str, token_pos: int, total_tokens: int) -> float:
        """Calculate steering weight based on strategy and token position."""
        position_frac = token_pos / max(total_tokens, 1)

        if strategy == "constant":
            return 1.0
        elif strategy == "initial_only":
            initial_tokens = config.get("initial_tokens", STEERING_APP_INITIAL_TOKENS)
            return 1.0 if token_pos < initial_tokens else 0.0
        elif strategy == "diminishing":
            rate = config.get("rate", STEERING_APP_DECAY_RATE)
            return math.exp(-rate * token_pos)
        elif strategy == "increasing":
            rate = config.get("rate", STEERING_APP_DECAY_RATE)
            return 1.0 - math.exp(-rate * token_pos)
        elif strategy == "gaussian":
            center = config.get("gaussian_center", STEERING_APP_GAUSSIAN_CENTER)
            width = config.get("gaussian_width", STEERING_APP_GAUSSIAN_WIDTH)
            return math.exp(-((position_frac - center) ** 2) / (2 * width ** 2))
        else:
            return 1.0

    # Component-targeted steering: dispatch to submodule hooks
    component = getattr(steering_obj.metadata, "extraction_component", "residual_stream")
    if component != "residual_stream":
        from wisent.core.primitives.models.component_steering import register_component_steering_hooks
        def _strategy_fn(token_pos: int) -> float:
            return get_steering_weight(steering_strategy, token_pos, max_new_tokens)
        register_component_steering_hooks(
            self.hf_model, steering_obj, self._hook_group,
            base_strength, _strategy_fn,
        )
        return

    name_to_index = {str(i + 1): i for i in range(len(self._layers))}

    for layer_idx in steering_obj.metadata.layers:
        layer_name = str(layer_idx)
        if layer_name not in name_to_index:
            continue

        idx = name_to_index[layer_name]
        layer_module = self._layers[idx]

        def _hook_factory(obj: BaseSteeringObject, layer: int, strength: float, counter: dict, strategy: str, max_tokens: int):
            def _hook(_mod: nn.Module, _inp: tuple, out: torch.Tensor | tuple) -> torch.Tensor | tuple:
                # Get current sequence length to detect new tokens
                if isinstance(out, tuple):
                    hs = out[0]
                else:
                    hs = out

                seq_len = hs.shape[1]

                # On first call, record prompt length
                if counter["prompt_len"] == 0:
                    counter["prompt_len"] = seq_len

                # Calculate token position (0 = first generated token)
                token_pos = seq_len - counter["prompt_len"]
                if token_pos < 0:
                    token_pos = 0

                # Get weight for this token position
                weight = get_steering_weight(strategy, token_pos, max_tokens)
                effective_strength = strength * weight

                # Skip steering if weight is 0
                if effective_strength == 0:
                    return out

                if isinstance(out, tuple):
                    steered = obj.apply_steering(hs, layer=layer, base_strength=effective_strength)
                    steered = steered.to(dtype=hs.dtype)  # enforce dtype match
                    return (steered,) + out[1:]
                else:
                    steered = obj.apply_steering(out, layer=layer, base_strength=effective_strength)
                    return steered.to(dtype=out.dtype)  # enforce dtype match
            return _hook

        handle = layer_module.register_forward_hook(
            _hook_factory(steering_obj, layer_idx, base_strength, token_counter, steering_strategy, max_new_tokens)
        )
        self._hook_group.add(handle)


def _encode_one(
    self,
    message: list[ChatMessage],
    add_generation_prompt: bool = True,
    enable_thinking: bool = False,
) -> dict[str, torch.Tensor]:
    """
    Encode a single input in chat format.

    arguments:
        messages:
            list of {'role': str, 'content': str} dicts (chat messages).
        add_generation_prompt:
            If True, append the model's generation prompt at the end.
        enable_thinking:
            If False, disable thinking/reasoning mode.

    returns:
        dict with 'input_ids' and 'attention_mask' tensors.
    """
    try:
        try:
            result = self.tokenizer.apply_chat_template(
                message, tokenize=True,
                add_generation_prompt=add_generation_prompt,
                enable_thinking=enable_thinking,
                return_tensors="pt",
            )
        except TypeError:
            result = self.tokenizer.apply_chat_template(
                message, tokenize=True,
                add_generation_prompt=add_generation_prompt,
                return_tensors="pt",
            )
        ids = extract_token_ids(result)
    except ValueError as e:
        raise ChatTemplateNotAvailableError(cause=e)
    return {
        "input_ids": ids,
        "attention_mask": torch.ones_like(ids),
    }


def _batch_encode(
    self,
    inputs: list[list[ChatMessage]],
    add_generation_prompt: bool = True,
    enable_thinking: bool = False,
) -> dict[str, torch.Tensor]:
    """
    Batch-encode a list of chat messages.

    arguments:
        inputs:
            list of chat messages (each a list of {'role','content'} dicts).
        add_generation_prompt:
            If True, append the model's generation prompt at the end of each.
        enable_thinking:
            If False, disable thinking/reasoning mode.

    returns:
        dict with batched 'input_ids' and 'attention_mask' tensors.
    """
    singles = []
    for item in inputs:
        singles.append(self._encode_one(item, add_generation_prompt=add_generation_prompt, enable_thinking=enable_thinking))

    batch = self.tokenizer.pad(singles, padding=True, return_tensors="pt")

    batch = {k: v.to(self.device) for k, v in batch.items()}

    return batch


def _extract_assistant_response(self, text: str) -> str:
    """
    Extract only the assistant's response from a decoded chat template output.

    arguments:
        text:
            Full decoded text from tokenizer.batch_decode

    returns:
        Extracted assistant response text
    """
    # Look for the assistant marker in the decoded text
    if "assistant" in text:
        response = text.split("assistant", 1)[1]
        response = response.strip()
    else:
        response = text

    # Remove empty thinking blocks that Qwen adds when enable_thinking=False
    response = re.sub(r'^<think>\s*</think>\s*', '', response)

    # Also remove thinking blocks with content (full thinking output)
    response = re.sub(r'^<think>.*?</think>\s*', '', response, flags=re.DOTALL)

    return response.strip()



"""Visualization of steering directions and their relationships.

Shows multiple steering directions as vectors, their angles, and how they
relate to each other geometrically.
"""
import numpy as np
import torch
from typing import Dict, List, Optional, Any, Tuple
from sklearn.decomposition import PCA
from wisent.core.utils.config_tools.constants import (
    ZERO_THRESHOLD,
    DIRECTION_ORTHOGONAL_THRESHOLD,
    DIRECTION_ALIGNED_THRESHOLD,
    DIRECTION_HIGH_SIMILARITY,
    DIRECTION_MODERATE_SIMILARITY,
    DIRECTION_LOW_SIMILARITY,
    N_COMPONENTS_2D,
    ANGLE_DEVIATION_DEGREES_THRESHOLD,
)


def compute_direction_angles(
    directions: List[torch.Tensor],
    names: Optional[List[str]] = None,
) -> Dict[str, Any]:
    """
    Compute pairwise angles and cosine similarities between directions.

    Args:
        directions: List of steering direction tensors
        names: Optional names for each direction

    Returns:
        Dict with cosine_matrix, angle_matrix (degrees), and summary stats.
    """
    n = len(directions)
    if names is None:
        names = [f"dir_{i}" for i in range(n)]

    # Normalize directions
    dirs_np = []
    for d in directions:
        d_np = d.cpu().numpy() if isinstance(d, torch.Tensor) else d
        norm = np.linalg.norm(d_np)
        dirs_np.append(d_np / norm if norm > ZERO_THRESHOLD else d_np)

    dirs_matrix = np.stack(dirs_np)

    # Cosine similarity matrix
    cosine_matrix = dirs_matrix @ dirs_matrix.T

    # Angle matrix (in degrees)
    cosine_clipped = np.clip(cosine_matrix, -1, 1)
    angle_matrix = np.degrees(np.arccos(cosine_clipped))

    # Extract off-diagonal for stats
    mask = ~np.eye(n, dtype=bool)
    off_diag_cos = cosine_matrix[mask]
    off_diag_ang = angle_matrix[mask]

    return {
        "cosine_matrix": cosine_matrix.tolist(),
        "angle_matrix_degrees": angle_matrix.tolist(),
        "names": names,
        "n_directions": n,
        "mean_cosine": float(off_diag_cos.mean()) if n > 1 else 1.0,
        "min_cosine": float(off_diag_cos.min()) if n > 1 else 1.0,
        "max_cosine": float(off_diag_cos.max()) if n > 1 else 1.0,
        "mean_angle_degrees": float(off_diag_ang.mean()) if n > 1 else 0.0,
        "min_angle_degrees": float(off_diag_ang.min()) if n > 1 else 0.0,
        "max_angle_degrees": float(off_diag_ang.max()) if n > 1 else 0.0,
        "orthogonal_pairs": int((np.abs(off_diag_cos) < DIRECTION_ORTHOGONAL_THRESHOLD).sum() // 2) if n > 1 else 0,
        "aligned_pairs": int((off_diag_cos > DIRECTION_ALIGNED_THRESHOLD).sum() // 2) if n > 1 else 0,
    }


def plot_direction_similarity_matrix(
    directions: List[torch.Tensor],
    names: Optional[List[str]] = None,
) -> Dict[str, Any]:
    """
    Create data for a heatmap of direction similarities.

    Returns data suitable for plotting with matplotlib/seaborn.
    """
    angles = compute_direction_angles(directions, names)

    return {
        "matrix": angles["cosine_matrix"],
        "labels": angles["names"],
        "title": "Steering Direction Cosine Similarities",
        "colorbar_label": "Cosine Similarity",
        "vmin": -1,
        "vmax": 1,
        "summary": {
            "mean_similarity": angles["mean_cosine"],
            "orthogonal_pairs": angles["orthogonal_pairs"],
            "aligned_pairs": angles["aligned_pairs"],
        },
    }


def plot_directions_in_pca_space(
    directions: List[torch.Tensor],
    pos_activations: Optional[torch.Tensor] = None,
    neg_activations: Optional[torch.Tensor] = None,
    names: Optional[List[str]] = None,
    n_components: int = N_COMPONENTS_2D,
) -> Dict[str, Any]:
    """
    Project directions to PCA space and visualize as arrows.

    If activations provided, fits PCA on activations and projects directions.
    Otherwise, fits PCA on the directions themselves.

    Args:
        directions: List of steering direction tensors
        pos_activations: Optional positive activations for PCA fitting
        neg_activations: Optional negative activations for PCA fitting
        names: Optional names for directions
        n_components: 2 or 3 for visualization

    Returns:
        Dict with projected directions and activation points for plotting.
    """
    n = len(directions)
    if names is None:
        names = [f"dir_{i}" for i in range(n)]

    # Stack directions
    dirs_np = np.stack([
        d.cpu().numpy() if isinstance(d, torch.Tensor) else d
        for d in directions
    ])

    # Fit PCA
    if pos_activations is not None and neg_activations is not None:
        pos_np = pos_activations.cpu().numpy() if isinstance(pos_activations, torch.Tensor) else pos_activations
        neg_np = neg_activations.cpu().numpy() if isinstance(neg_activations, torch.Tensor) else neg_activations
        X = np.vstack([pos_np, neg_np])
        pca = PCA(n_components=n_components)
        pca.fit(X)
        pos_proj = pca.transform(pos_np)
        neg_proj = pca.transform(neg_np)
    else:
        pca = PCA(n_components=n_components)
        pca.fit(dirs_np)
        pos_proj = None
        neg_proj = None

    # Project directions (as vectors from origin)
    dirs_proj = pca.transform(dirs_np)

    # Normalize for visualization (scale to unit length in projected space)
    norms = np.linalg.norm(dirs_proj, axis=1, keepdims=True)
    dirs_proj_normalized = dirs_proj / (norms + ZERO_THRESHOLD)

    return {
        "directions_projected": dirs_proj_normalized.tolist(),
        "directions_raw_projected": dirs_proj.tolist(),
        "names": names,
        "n_components": n_components,
        "explained_variance": pca.explained_variance_ratio_.tolist(),
        "pos_projected": pos_proj.tolist() if pos_proj is not None else None,
        "neg_projected": neg_proj.tolist() if neg_proj is not None else None,
        "title": f"Steering Directions in PC1-PC{n_components} Space",
    }


def compute_per_concept_directions(
    pos: torch.Tensor,
    neg: torch.Tensor,
    cluster_labels: List[int],
    n_concepts: int,
) -> Tuple[List[torch.Tensor], List[str]]:
    """
    Compute steering direction for each concept cluster.

    Args:
        pos: Positive activations
        neg: Negative activations
        cluster_labels: Cluster assignment for each sample
        n_concepts: Number of concepts

    Returns:
        Tuple of (directions list, names list).
    """
    labels = np.array(cluster_labels)
    directions = []
    names = []

    for concept_id in range(n_concepts):
        mask = labels == concept_id
        n_samples = mask.sum()

        if n_samples < 2:
            continue

        pos_concept = pos[mask]
        neg_concept = neg[mask]
        direction = pos_concept.mean(dim=0) - neg_concept.mean(dim=0)

        directions.append(direction)
        names.append(f"concept_{concept_id} (n={n_samples})")

    return directions, names


def visualize_concept_directions(
    pos: torch.Tensor,
    neg: torch.Tensor,
    cluster_labels: List[int],
    n_concepts: int,
) -> Dict[str, Any]:
    """
    Full visualization of per-concept steering directions.

    Returns:
        Dict with angles, similarity matrix, and PCA projection data.
    """
    directions, names = compute_per_concept_directions(
        pos, neg, cluster_labels, n_concepts
    )

    if len(directions) < 2:
        return {
            "n_directions": len(directions),
            "error": "Need at least 2 concept directions to visualize relationships",
        }

    angles = compute_direction_angles(directions, names)
    similarity_matrix = plot_direction_similarity_matrix(directions, names)
    pca_viz = plot_directions_in_pca_space(directions, pos, neg, names)

    return {
        "n_directions": len(directions),
        "names": names,
        "angles": angles,
        "similarity_matrix": similarity_matrix,
        "pca_visualization": pca_viz,
        "interpretation": _interpret_direction_relationships(angles),
    }


def _interpret_direction_relationships(angles: Dict[str, Any]) -> List[str]:
    """Generate interpretation of direction relationships."""
    interpretations = []

    mean_cos = angles["mean_cosine"]
    orthogonal = angles["orthogonal_pairs"]
    aligned = angles["aligned_pairs"]
    n = angles["n_directions"]

    if mean_cos > DIRECTION_HIGH_SIMILARITY:
        interpretations.append("Directions are highly aligned - may be same concept")
    elif mean_cos > DIRECTION_MODERATE_SIMILARITY:
        interpretations.append("Directions are moderately correlated")
    elif mean_cos > DIRECTION_LOW_SIMILARITY:
        interpretations.append("Directions are weakly correlated")
    elif mean_cos > -DIRECTION_LOW_SIMILARITY:
        interpretations.append("Directions are approximately orthogonal")
    else:
        interpretations.append("Directions are anti-correlated (opposing)")

    if orthogonal > 0:
        interpretations.append(f"{orthogonal} direction pairs are nearly orthogonal")

    if aligned > 0 and n > 2:
        interpretations.append(f"{aligned} direction pairs are nearly identical")

    if angles["max_angle_degrees"] - angles["min_angle_degrees"] > ANGLE_DEVIATION_DEGREES_THRESHOLD:
        interpretations.append("Direction angles vary widely - concepts are distinct")

    return interpretations



"""Helper functions for mixed benchmark sampling."""

import random
import logging
from typing import List, Dict, Any, Optional
from wisent.core.utils.config_tools.constants import SAMPLES_PER_BENCHMARK

logger = logging.getLogger(__name__)


def sample_benchmarks_by_tag(
    tag: str,
    samples_per_benchmark: int = SAMPLES_PER_BENCHMARK,
    max_benchmarks: Optional[int] = None,
    random_seed: Optional[int] = None
) -> Dict[str, List[Dict[str, Any]]]:
    """
    Convenience function to sample from all benchmarks with a specific tag.

    Args:
        tag: Tag to filter benchmarks (e.g., "coding")
        samples_per_benchmark: Number of samples from each benchmark
        max_benchmarks: Maximum number of benchmarks to sample from
        random_seed: Random seed for reproducibility

    Returns:
        Dictionary mapping benchmark names to their samples
    """
    from wisent.core.utils.services.benchmarks.mixed_benchmark_sampler import MixedBenchmarkSampler
    from wisent.core.utils.services.benchmarks.cache.managed_cached_benchmarks import get_managed_cache

    sampler = MixedBenchmarkSampler()

    # Get all benchmarks with the tag
    benchmarks = sampler.get_benchmarks_by_tag(tag)

    if max_benchmarks and len(benchmarks) > max_benchmarks:
        if random_seed is not None:
            random.seed(random_seed)
        benchmarks = random.sample(benchmarks, max_benchmarks)

    # Sample from each benchmark
    results = {}
    cache = get_managed_cache()

    for benchmark_name in benchmarks:
        try:
            samples = cache.get_task_samples(
                task_name=benchmark_name,
                limit=samples_per_benchmark,
                force_fresh=False
            )
            results[benchmark_name] = samples
            logger.info(f"Sampled {len(samples)} from {benchmark_name}")

        except Exception as e:
            logger.warning(f"Failed to sample from {benchmark_name}: {e}")
            continue

    return results


def create_mixed_contrastive_pair_set_impl(
    sampler,
    tags: List[str],
    total_samples: int,
    name: Optional[str] = None,
    **kwargs
):
    """
    Create a ContrastivePairSet from mixed benchmark samples.

    Args:
        sampler: MixedBenchmarkSampler instance
        tags: Tags to filter benchmarks
        total_samples: Number of samples to include
        name: Name for the pair set (auto-generated if not provided)
        **kwargs: Additional arguments for sample_mixed_dataset

    Returns:
        ContrastivePairSet ready for training
    """
    from wisent.core.primitives.contrastive_pairs import ContrastivePairSet

    # Sample mixed dataset
    train_samples, test_samples = sampler.sample_mixed_dataset(
        tags=tags,
        total_samples=total_samples,
        **kwargs
    )

    # Extract contrastive pairs
    all_samples = train_samples + test_samples
    contrastive_pairs = sampler.extract_contrastive_pairs_from_mixed_samples(all_samples)

    # Create name if not provided
    if name is None:
        name = f"mixed_{'_'.join(tags)}_{total_samples}_samples"

    # Create ContrastivePairSet
    return ContrastivePairSet.from_contrastive_pairs(
        name=name,
        contrastive_pairs=contrastive_pairs,
        task_type="mixed_benchmark"
    )



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from wisent.core.primitives.contrastive_pairs.core.pair import ContrastivePair
from wisent.core.primitives.contrastive_pairs.core.io.response import NegativeResponse, PositiveResponse
from wisent.extractors.lm_eval.atoms import LMEvalBenchmarkExtractor
from wisent.core.utils.cli.cli_logger import setup_logger, bind

if TYPE_CHECKING:
    from lm_eval.api.task import ConfigurableTask


__all__ = ["MathQAExtractor"]
_LOG = setup_logger(__name__)


class MathQAExtractor(LMEvalBenchmarkExtractor):
    """Extractor for MathQA benchmark."""

    def extract_contrastive_pairs(
        self,
        lm_eval_task_data: ConfigurableTask,
        limit: int | None = None,
    ) -> list[ContrastivePair]:
        """
        Build contrastive pairs from MathQA docs.

        Args:
            lm_eval_task_data: lm-eval task instance.
            limit: Optional maximum number of pairs to produce.

        Returns:
            A list of ContrastivePair objects.
        """
        log = bind(_LOG, task=getattr(lm_eval_task_data, "NAME", "unknown"))

        max_items = self._normalize_limit(limit)
        docs = self.load_docs(lm_eval_task_data, max_items)

        pairs: list[ContrastivePair] = []

        log.info("Extracting contrastive pairs", extra={"doc_count": len(docs)})

        for doc in docs:
            pair = self._extract_pair_from_doc(doc, lm_eval_task_data)
            if pair is not None:
                pairs.append(pair)
                if max_items is not None and len(pairs) >= max_items:
                    break

        if not pairs:
            task_name = getattr(lm_eval_task_data, "NAME", type(lm_eval_task_data).__name__)
            log.warning("No valid pairs extracted", extra={"task": task_name})

        return pairs

    def _extract_pair_from_doc(
        self, doc: dict[str, Any], task_data: Any = None
    ) -> ContrastivePair | None:
        """Convert a single doc into a ContrastivePair."""
        try:
            # mathqa format: Problem, options, correct
            problem = doc.get("Problem", "").strip()
            options_str = doc.get("options", "")
            correct_key = doc.get("correct", "").strip().lower()

            if not problem or not options_str or not correct_key:
                _LOG.debug("Skipping: missing problem, options, or correct answer")
                return None

            # Parse options string like "a ) 38 , b ) 27.675 , c ) 30 , d ) data inadequate , e ) none of these"
            options_dict = {}
            parts = options_str.split(',')
            for part in parts:
                part = part.strip()
                if ')' in part:
                    key, value = part.split(')', 1)
                    key = key.strip().lower()
                    value = value.strip()
                    options_dict[key] = value

            if correct_key not in options_dict:
                _LOG.debug(f"Skipping: correct key '{correct_key}' not in options")
                return None

            correct_answer = options_dict[correct_key]

            # Get an incorrect answer (any other option)
            incorrect_key = None
            for key in options_dict:
                if key != correct_key:
                    incorrect_key = key
                    break

            if incorrect_key is None:
                _LOG.debug("Skipping: no incorrect option available")
                return None

            incorrect_answer = options_dict[incorrect_key]

            # Format question
            formatted_question = f"Question: {problem}\nAnswer:"

            metadata = {
                "label": "mathqa",
                "source": getattr(task_data, "NAME", "mathqa"),
            }

            return self._build_pair(
                question=formatted_question,
                correct=correct_answer,
                incorrect=incorrect_answer,
                metadata=metadata,
            )

        except Exception as exc:
            _LOG.error(f"Error extracting pair from doc: {exc}", exc_info=True)
            return None

    @staticmethod
    def _build_pair(
        question: str,
        correct: str,
        incorrect: str,
        metadata: dict[str, Any] | None = None,
    ) -> ContrastivePair:
        """Build a ContrastivePair from question and responses."""
        positive_response = PositiveResponse(model_response=correct)
        negative_response = NegativeResponse(model_response=incorrect)
        return ContrastivePair(
            prompt=question,
            positive_response=positive_response,
            negative_response=negative_response,
            label=metadata.get("label") if metadata else None,
            metadata=metadata,
        )



"""Constants for benchmark processing."""

from __future__ import annotations

# Path to lm_eval tasks directory
LM_EVAL_TASKS_PATH = (
    "/opt/homebrew/Caskroom/miniforge/base/lib/python3.10/site-packages/lm_eval/tasks"
)

# Approved tags from skills.json and risks.json
APPROVED_SKILLS = [
    "coding",
    "mathematics",
    "long context",
    "creative writing",
    "general knowledge",
    "medical",
    "law",
    "science",
    "history",
    "tool use",
    "multilingual",
    "reasoning",
]

APPROVED_RISKS = [
    "harmfulness",
    "toxicity",
    "bias",
    "hallucination",
    "violence",
    "adversarial robustness",
    "sycophancy",
    "deception",
]

__all__ = ["LM_EVAL_TASKS_PATH", "APPROVED_SKILLS", "APPROVED_RISKS"]



"""ReFT evaluation: apply_reft_to_model, remove_reft, evaluate_reft."""
from __future__ import annotations
import gc
import json
from pathlib import Path
import torch
from wisent.core.utils.config_tools.constants import (
    COMPARISON_DEFAULT_BATCH_SIZE,
    COMPARISON_MAX_BATCH_SIZE,
    COMPARISON_NUM_PAIRS,
    COMPARISON_STEERING_LAYER,
    COMPARISON_STEERING_SCALES,
    DEFAULT_SPLIT_RATIO, JSON_INDENT,
)
from wisent.comparison.utils import (
    create_test_only_task, run_ll_evaluation, apply_steering_to_model, remove_steering,
)
from wisent.core.primitives.models.wisent_model import WisentModel

__all__ = ["apply_reft_to_model", "remove_reft", "evaluate_reft"]


def apply_reft_to_model(wisent_model: "WisentModel", reft_path: str | Path) -> None:
    """Apply a trained ReFT intervention to a WisentModel."""
    import pyreft
    reft_path = Path(reft_path)
    reft_model = pyreft.ReftModel.load(str(reft_path), wisent_model.hf_model)
    reft_model.set_device(wisent_model.device)
    model_dtype = next(wisent_model.hf_model.parameters()).dtype
    for k, v in reft_model.interventions.items():
        intervention = v[0] if isinstance(v, (list, tuple)) else v
        if hasattr(intervention, 'to'):
            intervention.to(dtype=model_dtype)

    class ReftModelWrapper(torch.nn.Module):
        def __init__(self, reft_model, original_model):
            super().__init__()
            self._reft_model = reft_model
            self._original_model = original_model
            self.config = original_model.config
            self.device = next(original_model.parameters()).device
        def forward(self, input_ids=None, attention_mask=None, **kwargs):
            base = {}
            if input_ids is not None:
                base["input_ids"] = input_ids
            if attention_mask is not None:
                base["attention_mask"] = attention_mask
            base.update(kwargs)
            _, outputs = self._reft_model(base=base)
            return outputs
        def __getattr__(self, name):
            if name in ("_reft_model", "_original_model", "config", "device",
                        "training", "_parameters", "_buffers", "_modules"):
                return super().__getattr__(name)
            return getattr(self._original_model, name)

    wrapper = ReftModelWrapper(reft_model, wisent_model.hf_model)
    wisent_model._original_model = wisent_model.hf_model
    wisent_model.hf_model = wrapper
    print(f"ReFT intervention loaded from {reft_path}")


def remove_reft(wisent_model: "WisentModel") -> None:
    """Remove/disable ReFT intervention from a WisentModel."""
    if hasattr(wisent_model, '_original_model'):
        wisent_model.hf_model = wisent_model._original_model
        del wisent_model._original_model
        print("ReFT intervention removed")
    else:
        print("No ReFT intervention to remove")


def evaluate_reft(
    model_name: str, reft_path: str | Path, task: str,
    train_ratio: float = DEFAULT_SPLIT_RATIO, device: str = "cuda:0",
    batch_size: int = COMPARISON_DEFAULT_BATCH_SIZE, max_batch_size: int = COMPARISON_MAX_BATCH_SIZE, limit: int | None = None,
    output_dir: str | Path = None,
    num_train_pairs: int | None = None, num_epochs: int | None = None,
    low_rank_dimension: int | None = None, intervention_layers: list[int] | None = None,
    learning_rate: float | None = None,
    with_steering: bool = False, steering_method: str = "caa",
    steering_layers: str = str(COMPARISON_STEERING_LAYER), steering_num_pairs: int = COMPARISON_NUM_PAIRS,
    steering_scales: list[float] | None = None, extraction_strategy: str = "mc_completion",
) -> dict:
    """Evaluate a trained ReFT intervention comparing base vs ReFT performance."""
    reft_path = Path(reft_path)
    if steering_scales is None:
        steering_scales = list(COMPARISON_STEERING_SCALES)
    print(f"\n{'='*60}\nCreating test task for: {task}\n{'='*60}")
    task_dict = create_test_only_task(task, train_ratio=train_ratio)
    print(f"\n{'='*60}\nLoading model: {model_name}\n{'='*60}")
    wisent_model = WisentModel(model_name=model_name, device=device)
    print(f"\n{'='*60}\nRunning BASE evaluation (no ReFT)\n{'='*60}")
    base_acc_ll = run_ll_evaluation(wisent_model, task_dict, task, limit)
    print(f"Base accuracy: {base_acc_ll:.4f}")
    print(f"\n{'='*60}\nApplying ReFT intervention from: {reft_path}\n{'='*60}")
    apply_reft_to_model(wisent_model, reft_path)
    print(f"\n{'='*60}\nRunning REFT evaluation\n{'='*60}")
    reft_acc_ll = run_ll_evaluation(wisent_model, task_dict, task, limit)
    print(f"ReFT accuracy: {reft_acc_ll:.4f}")
    results = {
        "task": task, "model": model_name, "reft_path": str(reft_path),
        "num_train_pairs": num_train_pairs, "num_epochs": num_epochs,
        "low_rank_dimension": low_rank_dimension, "intervention_layers": intervention_layers,
        "learning_rate": learning_rate, "train_ratio": train_ratio, "eval_limit": limit,
        "base_accuracy": base_acc_ll, "reft_accuracy": reft_acc_ll, "reft_diff": reft_acc_ll - base_acc_ll,
    }
    if with_steering:
        results = _eval_reft_with_steering(
            wisent_model, task, task_dict, limit, base_acc_ll, reft_acc_ll,
            steering_method, steering_layers, steering_num_pairs, steering_scales,
            extraction_strategy, device, results,
        )
    remove_reft(wisent_model)
    del wisent_model
    gc.collect()
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
    _print_reft_summary(task, model_name, reft_path, base_acc_ll, reft_acc_ll, with_steering, steering_method, results)
    if output_dir:
        output_dir = Path(output_dir) / model_name.replace("/", "_")
        output_dir.mkdir(parents=True, exist_ok=True)
        results_file = output_dir / f"{task}_reft_eval_results.json"
        with open(results_file, "w") as f:
            json.dump(results, f, indent=JSON_INDENT)
        print(f"\nResults saved to: {results_file}")
    return results


def _eval_reft_with_steering(wisent_model, task, task_dict, limit, base_acc_ll, reft_acc_ll,
                              steering_method, steering_layers, steering_num_pairs,
                              steering_scales, extraction_strategy, device, results):
    """Run ReFT + steering evaluation at multiple scales."""
    from wisent.core.weight_modification.trainers.steering_trainer import WisentSteeringTrainer
    from wisent.core.control.steering_methods import get_steering_method
    from wisent.core.primitives.model_interface.core.activations import ExtractionStrategy
    from wisent.core.primitives.contrastive_pairs.core.set import ContrastivePairSet
    from wisent.core.primitives.contrastive_pairs.core.pair import ContrastivePair
    from wisent.core.primitives.contrastive_pairs.core.io.response import PositiveResponse, NegativeResponse
    from wisent.comparison.utils import generate_contrastive_pairs
    pairs_data, pairs_file = generate_contrastive_pairs(task, steering_num_pairs)
    pairs = [
        ContrastivePair(
            prompt=p["prompt"],
            positive_response=PositiveResponse(model_response=p["positive_response"]["model_response"]),
            negative_response=NegativeResponse(model_response=p["negative_response"]["model_response"]),
        ) for p in pairs_data
    ]
    pair_set = ContrastivePairSet(pairs=pairs, name=f"{task}_reft_steering")
    steering_method_obj = get_steering_method(steering_method, device=device)
    strategy = ExtractionStrategy(extraction_strategy)
    trainer = WisentSteeringTrainer(model=wisent_model, pair_set=pair_set, steering_method=steering_method_obj)
    result = trainer.run(layers_spec=steering_layers, strategy=strategy, accept_low_quality_vector=True)
    steering_vectors = {k: v.cpu().float().tolist() for k, v in result.steered_vectors.to_dict().items() if v is not None}
    steering_data = {"steering_vectors": steering_vectors, "layers": list(steering_vectors.keys())}
    import os
    os.unlink(pairs_file)
    results["steering"] = {"method": steering_method, "layers": list(steering_vectors.keys()),
                           "num_pairs": steering_num_pairs, "extraction_strategy": extraction_strategy, "scales": {}}
    for scale in steering_scales:
        apply_steering_to_model(wisent_model, steering_data, scale=scale)
        steer_acc_ll = run_ll_evaluation(wisent_model, task_dict, task, limit)
        remove_steering(wisent_model)
        results["steering"]["scales"][str(scale)] = {
            "accuracy": steer_acc_ll, "diff_from_base": steer_acc_ll - base_acc_ll,
            "diff_from_reft": steer_acc_ll - reft_acc_ll,
        }
    return results


def _print_reft_summary(task, model_name, reft_path, base_acc_ll, reft_acc_ll, with_steering, steering_method, results):
    """Print ReFT results summary."""
    print(f"\n{'='*50}\nRESULTS SUMMARY\n{'='*50}")
    print(f"Task: {task}\nModel: {model_name}\nReFT: {reft_path}")
    print(f"{'-'*50}\n{'Method':<30} {'Accuracy':<12} {'Diff':<10}\n{'-'*50}")
    print(f"{'Base':<30} {base_acc_ll:<12.4f} {'':<10}")
    print(f"{'ReFT':<30} {reft_acc_ll:<12.4f} {reft_acc_ll - base_acc_ll:+.4f}")
    if with_steering:
        for scale, res in results["steering"]["scales"].items():
            label = f"ReFT+{steering_method.upper()}@{scale}"
            print(f"{label:<30} {res['accuracy']:<12.4f} {res['diff_from_base']:+.4f}")
    print(f"{'='*50}")
